
import io
import enum
from typing import List
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/89fa11de-5275-4106-9cf1-e5aa7709436c
class TS_UD_SC_NET:
	def __init__(self):
		self.type:TS_UD_TYPE = TS_UD_TYPE.SC_NET
		self.length:int = None
		self.MCSChannelId:int = None
		self.channelCount:int = None
		self.channelIdArray:List[int] = []
		self.Pad = None #optional
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = self.MCSChannelId.to_bytes(4, byteorder='little', signed = False)
		t += len(self.channelIdArray).to_bytes(2, byteorder='little', signed = False)
		for cd in self.channelIdArray:
			t += cd.to_bytes(2, byteorder='little', signed = False)
		if len(self.channelIdArray) % 2 != 0:
			t+= b'\x00\x00' # padding to multiple of 4
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_SC_NET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UD_SC_NET()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.MCSChannelId = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.channelCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		for _ in range(msg.channelCount):
			msg.channelIdArray.append(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		if msg.channelCount % 2 != 0:
			msg.Pad = buff.read(2)
		return msg

	def __repr__(self):
		t = '==== TS_UD_SC_NET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum
from typing import List

class SURFCMD(enum.Enum):
	CMDTYPE_SET_SURFACE_BITS = 0x0001 #Indicates a Set Surface Bits Command (section 2.2.9.2.1).
	CMDTYPE_FRAME_MARKER = 0x0004 #Indicates a Frame Marker Command (section 2.2.9.2.3).
	CMDTYPE_STREAM_SURFACE_BITS = 0x0006 #Indicates a Stream Surface Bits Command (section 2.2.9.2.2).

class TS_COMPRESSED_BITMAP_HEADER_EX:
	def __init__(self):
		self.highUniqueId:int = None
		self.lowUniqueId:int = None
		self.tmMilliseconds:int = None
		self.tmSeconds:int = None

	def to_bytes(self):
		t = self.highUniqueId.to_bytes(4, byteorder='little', signed=False)
		t += self.lowUniqueId.to_bytes(4, byteorder='little', signed=False)
		t += self.tmMilliseconds.to_bytes(8, byteorder='little', signed=False)
		t += self.tmSeconds.to_bytes(8, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_COMPRESSED_BITMAP_HEADER_EX.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_COMPRESSED_BITMAP_HEADER_EX()
		msg.highUniqueId = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		msg.lowUniqueId = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		msg.tmMilliseconds = int.from_bytes(buff.read(8), byteorder='little', signed=False)
		msg.tmSeconds = int.from_bytes(buff.read(8), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_COMPRESSED_BITMAP_HEADER_EX ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class BITMAP_EX_FLAG(enum.IntFlag):
	EX_COMPRESSED_BITMAP_HEADER_PRESENT = 0x01 #Indicates that the optional exBitmapDataHeader field is present.

class TS_BITMAP_DATA_EX:
	def __init__(self):
		self.bpp:int = None
		self.flags:BITMAP_EX_FLAG = None
		self.reserved:int = None
		self.codecID:int = None
		self.width:int = None
		self.height:int = None
		self.bitmapDataLength:int = None
		self.exBitmapDataHeader:TS_COMPRESSED_BITMAP_HEADER_EX = None
		self.bitmapData:int = None

	def to_bytes(self):
		t = self.bpp.to_bytes(1, byteorder='little', signed=False)
		t += self.flags.to_bytes(1, byteorder='little', signed=False)
		t += self.reserved.to_bytes(1, byteorder='little', signed=False)
		t += self.codecID.to_bytes(1, byteorder='little', signed=False)
		t += self.width.to_bytes(2, byteorder='little', signed=False)
		t += self.height.to_bytes(2, byteorder='little', signed=False)
		t += len(self.bitmapData).to_bytes(4, byteorder='little', signed=False)
		if BITMAP_EX_FLAG.EX_COMPRESSED_BITMAP_HEADER_PRESENT in self.flags:
			t += self.exBitmapDataHeader.to_bytes()
		t += self.bitmapData
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAP_DATA_EX.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAP_DATA_EX()
		msg.bpp = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		msg.flags = BITMAP_EX_FLAG(int.from_bytes(buff.read(1), byteorder='little', signed=False))
		msg.reserved = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		msg.codecID = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		msg.width = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.height = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.bitmapDataLength = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		if BITMAP_EX_FLAG.EX_COMPRESSED_BITMAP_HEADER_PRESENT in msg.flags:
			msg.exBitmapDataHeader = TS_COMPRESSED_BITMAP_HEADER_EX.from_buffer(buff)
		msg.bitmapData = buff.read(msg.bitmapDataLength)
		return msg

	def __repr__(self):
		t = '==== TS_BITMAP_DATA_EX ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



class TS_SURFCMD_SET_SURF_BITS:
	def __init__(self):
		self.destLeft:int = None
		self.destTop:int = None
		self.destRight:int = None
		self.destBottom:int = None
		self.bitmapData:TS_BITMAP_DATA_EX = None

	def to_bytes(self):
		t = self.destLeft.to_bytes(2, byteorder='little', signed=False)
		t += self.destTop.to_bytes(2, byteorder='little', signed=False)
		t += self.destRight.to_bytes(2, byteorder='little', signed=False)
		t += self.destBottom.to_bytes(2, byteorder='little', signed=False)
		t += self.bitmapData.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SURFCMD_SET_SURF_BITS.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SURFCMD_SET_SURF_BITS()
		msg.destLeft = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destTop = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destRight = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destBottom = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.bitmapData = TS_BITMAP_DATA_EX.from_buffer(buff)
		return msg

	def __repr__(self):
		t = '==== TS_SURFCMD_SET_SURF_BITS ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class SURFACECMD_FRAMEACTION(enum.Enum):
	BEGIN = 0x0000 #Indicates the start of a new frame.
	END = 0x0001 #Indicates the end of the current frame.

class TS_FRAME_MARKER:
	def __init__(self):
		self.frameAction:SURFACECMD_FRAMEACTION = None
		self.frameId:int = None

	def to_bytes(self):
		t = self.frameAction.value.to_bytes(2, byteorder='little', signed=False)
		t += self.frameId.to_bytes(4, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FRAME_MARKER.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FRAME_MARKER()
		msg.frameAction = SURFACECMD_FRAMEACTION(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		msg.frameId = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_FRAME_MARKER ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_SURFCMD_STREAM_SURF_BITS:
	def __init__(self):
		self.destLeft:int = None
		self.destTop:int = None
		self.destRight:int = None
		self.destBottom:int = None
		self.bitmapData:TS_BITMAP_DATA_EX = None

	def to_bytes(self):
		t = self.destLeft.to_bytes(2, byteorder='little', signed=False)
		t += self.destTop.to_bytes(2, byteorder='little', signed=False)
		t += self.destRight.to_bytes(2, byteorder='little', signed=False)
		t += self.destBottom.to_bytes(2, byteorder='little', signed=False)
		t += self.bitmapData.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SURFCMD_STREAM_SURF_BITS.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SURFCMD_STREAM_SURF_BITS()
		msg.destLeft = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destTop = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destRight = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destBottom = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.bitmapData = TS_BITMAP_DATA_EX.from_buffer(buff)
		return msg

	def __repr__(self):
		t = '==== TS_SURFCMD_STREAM_SURF_BITS ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_SURFCMD:
	def __init__(self):
		self.cmdType:SURFCMD = None
		self.cmdData = None

	def to_bytes(self):
		t = self.cmdType.value.to_bytes(2, byteorder='little', signed=False)
		t += self.cmdData.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SURFCMD.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SURFCMD()
		msg.cmdType = SURFCMD(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		if msg.cmdType == SURFCMD.CMDTYPE_SET_SURFACE_BITS:
			msg.cmdData = TS_SURFCMD_SET_SURF_BITS.from_buffer(buff)
		elif msg.cmdType == SURFCMD.CMDTYPE_FRAME_MARKER:
			msg.cmdData = TS_FRAME_MARKER.from_buffer(buff)
		elif msg.cmdType == SURFCMD.CMDTYPE_STREAM_SURFACE_BITS:
			msg.cmdData = TS_SURFCMD_STREAM_SURF_BITS.from_buffer(buff)
		return msg

	def __repr__(self):
		t = '==== TS_SURFCMD ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_FP_SURFCMDS:
	def __init__(self):
		self.surfaceCommands:List[TS_SURFCMD] = []

	def to_bytes(self):
		t = b''
		for cmd in self.surfaceCommands:
			t = cmd.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FP_SURFCMDS.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FP_SURFCMDS()
		
		return msg

	def __repr__(self):
		t = '==== TS_FP_SURFCMDS ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
import enum
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/9269d58a-3d85-48a2-942a-bb0bbe5a55aa
class TS_UD_SC_MCS_MSGCHANNEL:
	def __init__(self):
		self.length:int = None
		self.type:TS_UD_TYPE = TS_UD_TYPE.CS_MCS_MSGCHANNEL
		self.MCSChannelID:int = None
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = self.MCSChannelID.to_bytes(2, byteorder='little', signed = False)
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_SC_MCS_MSGCHANNEL.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UD_SC_MCS_MSGCHANNEL()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.MCSChannelID = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_UD_SC_MCS_MSGCHANNEL ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

class TS_RAIL_LEVEL(enum.IntFlag):
	SUPPORTED = 0x1 #Set to 1 if the client/server is capable of supporting Remote Programs; set to 0 otherwise.
	DOCKED_LANGBAR_SUPPORTED = 0x2 #Set to 1 if the client/server is capable of supporting Docked Language Bar for Remote Programs; set to 0 otherwise. This flag MUST be set to 0 if TS_RAIL_LEVEL_SUPPORTED is 0.
	SHELL_INTEGRATION_SUPPORTED = 0x4 #Set to 1 if the client/server is capable of supporting extended shell integration like tabbed windows and overlay icons for Remote Programs; set to 0 otherwise. This flag MUST be set to 0 if TS_RAIL_LEVEL_SUPPORTED is 0.
	LANGUAGE_IME_SYNC_SUPPORTED = 0x8 #Set to 1 if the client/server is capable of supporting syncing language/IME changes for Remote Programs; set to 0 otherwise. This flag MUST be set to 0 if TS_RAIL_LEVEL_SUPPORTED is 0.
	SERVER_TO_CLIENT_IME_SYNC_SUPPORTED = 0x10 #Set to 1 if the client/server is capable of supporting syncing IME changes originating at the server for Remote Programs; set to 0 otherwise. This flag MUST be set to 0 if TS_RAIL_LEVEL_SUPPORTED is 0.
	HIDE_MINIMIZED_APPS_SUPPORTED = 0x20 #Set to 1 if the client/server supports hiding minimized windows of Remote Programs on the server; set to 0 otherwise. This flag MUST be set to 0 if TS_RAIL_LEVEL_SUPPORTED is 0.
	WINDOW_CLOAKING_SUPPORTED = 0x40 #Set to 1 if the client/server supports syncing per-window cloak state changes originating on the client for Remote Programs; set to 0 otherwise. This flag MUST be set to 0 if TS_RAIL_LEVEL_SUPPORTED is 0.
	HANDSHAKE_EX_SUPPORTED = 0x80 #Set to 1 if the client/server supports the HandshakeEx PDU (section 2.2.2.2.3); set to 0 otherwise. This flag MUST be set to 0 if TS_RAIL_LEVEL_SUPPORTED is 0.

class TS_REMOTE_PROGRAMS_CAP_SET:
	def __init__(self):
		self.RailSupportLevel:TS_RAIL_LEVEL = None

	def to_bytes(self):
		t = self.RailSupportLevel.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_REMOTE_PROGRAMS_CAP_SET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_REMOTE_PROGRAMS_CAP_SET()
		msg.RailSupportLevel = TS_RAIL_LEVEL(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== TS_REMOTE_PROGRAMS_CAP_SET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

class CHANNEL_FLAG(enum.IntFlag):
	CHANNEL_FLAG_FIRST = 0x00000001 #Indicates that the chunk is the first in a sequence.
	CHANNEL_FLAG_LAST = 0x00000002 #Indicates that the chunk is the last in a sequence.
	CHANNEL_FLAG_SHOW_PROTOCOL = 0x00000010 #The Channel PDU Header MUST be visible to the application endpoint (section 2.2.1.3.4.1).
	CHANNEL_FLAG_SUSPEND = 0x00000020 #All virtual channel traffic MUST be suspended. This flag is only valid in server-to-client virtual channel traffic. It MUST be ignored in client-to-server data.
	CHANNEL_FLAG_RESUME = 0x00000040 #All virtual channel traffic MUST be resumed. This flag is only valid in server-to-client virtual channel traffic. It MUST be ignored in client-to-server data.
	CHANNEL_FLAG_SHADOW_PERSISTENT = 0x00000080 #This flag is unused and its value MUST be ignored by the client and server.
	CHANNEL_PACKET_COMPRESSED = 0x00200000 #The virtual channel data is compressed. This flag is equivalent to MPPC bit C (for more information see [RFC2118] section 3.1).
	CHANNEL_PACKET_AT_FRONT = 0x00400000 #The decompressed packet MUST be placed at the beginning of the history buffer. This flag is equivalent to MPPC bit B (for more information see [RFC2118] section 3.1).
	CHANNEL_PACKET_FLUSHED = 0x00800000 #The decompressor MUST reinitialize the history buffer (by filling it with zeros) and reset the HistoryOffset to zero. After it has been reinitialized, the entire history buffer is immediately regarded as valid. This flag is equivalent to MPPC bit A (for more information see [RFC2118] section 3.1). If the CHANNEL_PACKET_COMPRESSED (0x00200000) flag is also present, then the CHANNEL_PACKET_FLUSHED flag MUST be processed first.
	CompressionTypeMask = 0x000F0000 #Indicates the compression package which was used to compress the data. See the discussion which follows this table for a list of compression packages.

class CHANNEL_PDU_HEADER:
	def __init__(self):
		self.length:int = None
		self.flags:CHANNEL_FLAG = None
		self.data:bytes = None

	def to_bytes(self):
		t = self.length.to_bytes(4, byteorder='little', signed = False)
		t += self.flags.to_bytes(4, byteorder='little', signed = False)
		t += self.data
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CHANNEL_PDU_HEADER.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CHANNEL_PDU_HEADER()
		msg.length = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.flags = CHANNEL_FLAG(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.data = buff.read(msg.length)
		return msg

	@staticmethod
	def serialize_packet(flags:CHANNEL_FLAG, data, length = None):
		hdr = CHANNEL_PDU_HEADER()
		hdr.length = len(data)
		if length is not None:
			hdr.length = length
		hdr.data = data
		hdr.flags = flags
		return hdr

	def __repr__(self):
		t = '==== CHANNEL_PDU_HEADER ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
from aardwolf.extensions.RDPEDYC.protocol import DYNVC_CMD, dynvc_header_to_bytes, dynvc_header_from_buff

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/69377767-56a6-4ab8-996b-7758676e9261
class DYNVC_DATA_FIRST:
	def __init__(self):
		self.cbid:int = 4
		self.len:int   = 4
		self.cmd:DYNVC_CMD = DYNVC_CMD.DATA_FIRST
		self.ChannelId:int   = 0
		self.Length:int = None
		self.Data:bytes = None

	@staticmethod
	def from_bytes(data: bytes):
		return DYNVC_DATA_FIRST.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		hdr_start = buff.tell()
		msg = DYNVC_DATA_FIRST()
		msg.cbid, msg.len, msg.cmd = dynvc_header_from_buff(buff, cbid_mod=True, sp_mod = True)
		msg.ChannelId = int.from_bytes(buff.read(msg.cbid), byteorder='little', signed=False)
		msg.Length = int.from_bytes(buff.read(msg.len), byteorder='little', signed=False)
		hdr_end = buff.tell()
		hdrsize = hdr_end - hdr_start
		if hdrsize+msg.Length < 1600:
			msg.Data = buff.read(msg.Length)
		else:
			msg.Data = buff.read(1600 - hdrsize)

		return msg

	def to_bytes(self):
		t = b''
		t += dynvc_header_to_bytes(self.cbid, self.len, self.cmd, cbid_mod=True, sp_mod = True)
		t += self.ChannelId.to_bytes(self.cbid, byteorder='little', signed=False)
		t += self.Length.to_bytes(self.len, byteorder='little', signed=False)
		t += self.Data
		return t

	@staticmethod
	def chunk_data(data, channel_id):
		chunksize = 1600-9
		if len(data) < chunksize:
			msg = DYNVC_DATA()
			msg.cbid = 4
			msg.ChannelId = channel_id
			msg.Data = data
			yield msg.to_bytes()
		else:
			msg = DYNVC_DATA_FIRST()
			msg.cbid = 4
			msg.len  = 4
			msg.cmd  = DYNVC_CMD.DATA_FIRST
			msg.ChannelId = channel_id
			msg.Data = data[:chunksize]
			msg.Length = len(data)
			yield msg.to_bytes()

			i = chunksize
			while i < len(data):
				chunk = data[i:i+chunksize]
				msg = DYNVC_DATA()
				msg.cbid = 4
				msg.ChannelId = channel_id
				msg.Data = chunk
				yield msg.to_bytes()
				i += chunksize
			
	def __str__(self):
		t = ''
		for k in self.__dict__:
			if k == 'Data' and len(self.Data) > 100:
				res = 'Data: %s...(%s)\r\n' % (self.Data[50], len(self.Data))
			else:
				res = '%s: %s\r\n' % (k, self.__dict__[k])
			t += res
		return t

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/15b59886-db44-47f1-8da3-47c8fcd82803
class DYNVC_DATA:
	def __init__(self):
		self.cbid:int = None
		self.sp:int   = 0
		self.cmd:DYNVC_CMD = DYNVC_CMD.DATA
		self.ChannelId:int   = 0
		self.Data:bytes = None

	@staticmethod
	def from_bytes(data: bytes):
		return DYNVC_DATA.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		hdr_start = buff.tell()
		msg = DYNVC_DATA()
		msg.cbid, msg.sp, msg.cmd = dynvc_header_from_buff(buff, cbid_mod=True, sp_mod = True)
		msg.ChannelId = int.from_bytes(buff.read(msg.cbid), byteorder='little', signed=False)
		hdr_end = buff.tell()
		hdrsize = hdr_end - hdr_start
		msg.Data = buff.read(1600 - hdrsize)
		return msg

	def to_bytes(self):
		t = b''
		t += dynvc_header_to_bytes(self.cbid, self.sp, self.cmd, cbid_mod=True, sp_mod = True)
		t += self.ChannelId.to_bytes(self.cbid, byteorder='little', signed=False)
		t += self.Data
		return t


	def __str__(self):
		t = ''
		for k in self.__dict__:
			if k == 'Data' and len(self.Data) > 100:
				res = 'Data: %s...(%s)\r\n' % (self.Data[50], len(self.Data))
			else:
				res = '%s: %s\r\n' % (k, self.__dict__[k])
			t += res
		return t



import io
from typing import List
import enum
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, ChannelOption

class CHANNEL_DEF:
	def __init__(self):
		self.name:str = None
		self.options:ChannelOption = None
		
	def to_bytes(self):
		name = self.name.encode('ascii')
		name = name.ljust(8,b'\x00')
		t = name
		t += self.options.to_bytes(4, byteorder='big', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CHANNEL_DEF.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CHANNEL_DEF()
		msg.name = buff.read(8).decode('ascii').replace('\x00','')
		msg.options = ChannelOption(int.from_bytes(buff.read(4), byteorder='big', signed = False))
		return msg

	def __repr__(self):
		t = '==== CHANNEL_DEF ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/6b58e11e-a32b-4903-b736-339f3cfe46ec
class TS_UD_CS_NET:
	def __init__(self):
		self.type:TS_UD_TYPE = TS_UD_TYPE.CS_NET
		self.length:int = None
		self.channelCount:int = None
		self.channelDefArray:List[CHANNEL_DEF] = []
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = len(self.channelDefArray).to_bytes(4, byteorder='little', signed = False)
		for cd in self.channelDefArray:
			t += cd.to_bytes()
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_CS_NET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UD_CS_NET()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.channelCount = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		for _ in range(msg.channelCount):
			msg.channelDefArray.append(CHANNEL_DEF.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== TS_UD_CS_NET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


MCSPDU_ver_2 = """
MCS-PROTOCOL DEFINITIONS  ::= 
 
BEGIN 
 
--  Part 1: Fundamental MCS types 
 
ChannelId  ::=  INTEGER (0..65535)   --  range is 16 bits 
 
StaticChannelId  ::=  ChannelId (1..1000)  --  those known permanently 
 
DynamicChannelId  ::=  ChannelId (1001..65535) --  those created and deleted 
 
UserId  ::=  ChannelId --DynamicChannelId      created by Attach-User 
        --  deleted by Detach-User 
 
PrivateChannelId  ::=  DynamicChannelId  --  created by Channel-Convene 
        --  deleted by Channel-Disband 
 
AssignedChannelId  ::=  DynamicChannelId  --  created by Channel-Join zero 
        --  deleted by last Channel-Leave 
 
TokenId  ::=  INTEGER (1..65535)   --  all are known permanently 
 
TokenStatus  ::=  ENUMERATED 
{ 
 notInUse    (0), 
 selfGrabbed     (1), 
 otherGrabbed   (2), 
 selfInhibited    (3), 
 otherInhibited   (4), 
 selfRecipient    (5), 
 elfGiving    (6), 
 otherGiving     (7) 
} 
 
DataPriority  ::=  ENUMERATED 
{ 
 top     (0), 
 high     (1), 
 medium    (2), 
 low     (3) 
} 
 
Segmentation  ::=  BIT STRING 
{ 
 begin      (0), 
 end     (1) 
} (SIZE (2)) 
 
DomainParameters  ::=  SEQUENCE 
{ 
 maxChannelIds   INTEGER (0..MAX), 
      --  a limit on channel ids in use, 
      --  static + user id + private + assigned 
 maxUserIds     INTEGER (0..MAX), 
      --  a sublimit on user id channels alone 
 maxTokenIds   INTEGER (0..MAX), 
      --  a limit on token ids in use 
      --  grabbed + inhibited + giving + ungivable + given 
 numPriorities   INTEGER (0..MAX), 
      --  the number of TCs in an MCS connection 
 minThroughput   INTEGER (0..MAX), 
      --  the enforced number of octets per second 
 maxHeight    INTEGER  (0..MAX), 
      --  a limit on the height of a provider 
 maxMCSPDUsize   INTEGER  (0..MAX), 
      --  an octet limit on domain MCSPDUs 
 protocolVersion   INTEGER (0..MAX) 
} 

--  Part 2: Connect provider 
 
Connect-Initial  ::=  [APPLICATION 101] IMPLICIT SEQUENCE 
{ 
 callingDomainSelector  OCTET STRING, 
 calledDomainSelector  OCTET STRING, 
 upwardFlag     BOOLEAN, 
      --  TRUE if called provider is higher 
 targetParameters   DomainParameters, 
 minimumParameters  DomainParameters, 
 maximumParameters  DomainParameters, 
 userData    OCTET STRING 
} 
 
Connect-Response  ::=  [APPLICATION 102] IMPLICIT SEQUENCE 
{ 
 result      Result, 
 calledConnectId   INTEGER (0..MAX), 
      --  assigned by the called provider 
      --  to identify additional TCs of  
      --  the same MCS connection 
 domainParameters    DomainParameters, 
 userData    OCTET STRING 
} 

Connect-Additional  ::=  [APPLICATION 103] IMPLICIT SEQUENCE 
{ 
 calledConnectId   INTEGER (0..MAX), 
 dataPriority     DataPriority 
} 
 
Connect-Result  ::=  [APPLICATION 104] IMPLICIT SEQUENCE 
{ 
 result      Result 
} 
 
--  Part 3: Merge domain 
 
PlumbDomainIndication  ::=  [APPLICATION 0] IMPLICIT SEQUENCE 
{ 
 heightLimit    INTEGER (0..MAX) 
      --  a restriction on the MCSPDU receiver 
} 
 
ErectDomainRequest  ::=  [APPLICATION 1] IMPLICIT SEQUENCE 
{ 
 subHeight    INTEGER (0..MAX), 
      --  height in domain of the MCSPDU transmitter 
 subInterval    INTEGER (0..MAX) 
      --  its throughput enforcement interval in milliseconds 
} 
 
ChannelAttributes  ::=  CHOICE 
{ 
 static      [0] IMPLICIT SEQUENCE 
 { 
  channelId    StaticChannelId 
       --  joined is implicitly TRUE 
 }, 
 
 userId     [1] IMPLICIT SEQUENCE 
 { 
  joined     BOOLEAN, 
       --  TRUE if user is joined to its user id 
  userId     UserId 
 }, 
 
 private    [2] IMPLICIT SEQUENCE 
 { 
  joined     BOOLEAN, 
       --  TRUE if channel id is joined below 
  channelId    PrivateChannelId, 
  manager    UserId, 
  admitted    SET OF UserId 
       --  may span multiple MergeChannelsRequest 
 }, 
 
 assigned    [3] IMPLICIT SEQUENCE 
 { 
  channelId    AssignedChannelId 
       --  joined is implicitly TRUE 
 } 
}
MergeChannelsRequest  ::=  [APPLICATION 2] IMPLICIT SEQUENCE 
{ 
 mergeChannels   SET OF ChannelAttributes, 
 purgeChannelIds   SET OF ChannelId 
} 
 
MergeChannelsConfirm  ::=  [APPLICATION 3] IMPLICIT SEQUENCE 
{ 
 mergeChannels   SET OF ChannelAttributes, 
 purgeChannelIds   SET OF ChannelId 
} 
 
PurgeChannelsIndication  ::=  [APPLICATION 4] IMPLICIT SEQUENCE 
{ 
 detachUserIds   SET OF UserId, 
      --  purge user id channels 
 purgeChannelIds   SET OF ChannelId 
      --  purge other channels 
} 
 
TokenAttributes  ::=  CHOICE 
{ 
 grabbed    [0] IMPLICIT SEQUENCE 
 { 
  tokenId    TokenId, 
  grabber    UserId 
 }, 
 inhibited    [1] IMPLICIT SEQUENCE 
 { 
  tokenId    TokenId, 
  inhibitors    SET OF UserId 
       --  may span multiple MergeTokensRequest 
 }, 
 
 giving     [2] IMPLICIT SEQUENCE 
 { 
  tokenId    TokenId, 
  grabber    UserId, 
  recipient    UserId 
 }, 
 
 ungivable    [3] IMPLICIT SEQUENCE 
 { 
  tokenId    TokenId, 

  grabber    UserId 
       --  recipient has since detached 
 }, 
 
 given      [4] IMPLICIT SEQUENCE 
 { 
  tokenId    TokenId, 
  recipient    UserId 
       --  grabber released or detached 
 } 
} 
 
MergeTokensRequest  ::=  [APPLICATION 5] IMPLICIT SEQUENCE 
{ 
 mergeTokens   SET OF TokenAttributes, 
 purgeTokenIds   SET OF TokenId 
 } 
 
MergeTokensConfirm  ::=  [APPLICATION 6] IMPLICIT SEQUENCE 
{ 
 mergeTokens   SET OF TokenAttributes, 
 purgeTokenIds   SET OF TokenId 
} 
 
PurgeTokensIndication  ::=  [APPLICATION 7] IMPLICIT SEQUENCE 
{ 
 purgeTokenIds   SET OF TokenId 
} 
--  Part 4: Disconnect provider 
 
DisconnectProviderUltimatum  ::=  [APPLICATION 8] IMPLICIT SEQUENCE 
{ 
 reason    Reason 
} 
 
RejectMCSPDUUltimatum  ::=  [APPLICATION 9] IMPLICIT SEQUENCE 
{ 
 diagnostic    Diagnostic, 
 initialOctets     OCTET STRING 
} 
 
--  Part 5: Attach/Detach user 
 
AttachUserRequest  ::=  [APPLICATION 10] IMPLICIT SEQUENCE 
{ 
} 
 
AttachUserConfirm  ::=  [APPLICATION 11] IMPLICIT SEQUENCE 
{ 
 result      Result, 
 initiator    UserId OPTIONAL 
} 
 
DetachUserRequest  ::=  [APPLICATION 12] IMPLICIT SEQUENCE 
{ 
 reason    Reason, 
 userIds    SET OF UserId 
} 
 
DetachUserIndication  ::=  [APPLICATION 13] IMPLICIT SEQUENCE 
{ 
 reason    Reason, 
 userIds    SET OF UserId 
} 
 
--  Part 6: Channel management 
 
ChannelJoinRequest  ::=  [APPLICATION 14] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    ChannelId 
      --  may be zero 
}
ChannelJoinConfirm  ::=  [APPLICATION 15] IMPLICIT SEQUENCE 
{ 
 result      Result, 
 initiator    UserId, 
 requested    ChannelId, 
      --  may be zero 
 channelId    ChannelId OPTIONAL 
} 
 
ChannelLeaveRequest  ::=  [APPLICATION 16] IMPLICIT SEQUENCE 
{ 
 channelIds    SET OF ChannelId 
} 
 
ChannelConveneRequest  ::=  [APPLICATION 17] IMPLICIT SEQUENCE 
{ 
 initiator    UserId 
} 
 
ChannelConveneConfirm  ::=  [APPLICATION 18] IMPLICIT SEQUENCE 
{ 
 result      Result, 
 initiator    UserId, 
 channelId    PrivateChannelId OPTIONAL 
} 
 
ChannelDisbandRequest  ::=  [APPLICATION 19] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    PrivateChannelId 
} 
 
ChannelDisbandIndication  ::=  [APPLICATION 20] IMPLICIT SEQUENCE 
{ 
 channelId    PrivateChannelId 
} 
 
ChannelAdmitRequest  ::=  [APPLICATION 21] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    PrivateChannelId, 
 userIds    SET OF UserId 
} 
 
ChannelAdmitIndication  ::=  [APPLICATION 22] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    PrivateChannelId, 
 userIds    SET OF UserId 
} 
 
ChannelExpelRequest  ::=  [APPLICATION 23] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    PrivateChannelId, 
 userIds    SET OF UserId 
}


 ChannelExpelIndication  ::=  [APPLICATION 24] IMPLICIT SEQUENCE 
{ 
 channelId    PrivateChannelId, 
 userIds    SET OF UserId 
} 
 
--  Part 7: Data transfer 
 
SendDataRequest  ::=  [APPLICATION 25] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    ChannelId, 
 dataPriority     DataPriority, 
 segmentation   Segmentation, 
 userData    OCTET STRING 
} 
 
SendDataIndication  ::=  [APPLICATION 26] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    ChannelId, 
 dataPriority     DataPriority, 
 segmentation   Segmentation, 
 userData    OCTET STRING 
} 
 
UniformSendDataRequest  ::=  [APPLICATION 27] IMPLICIT SEQUENCE 
{ 

 initiator    UserId, 
 channelId    ChannelId, 
 dataPriority     DataPriority, 
 segmentation   Segmentation, 
 userData    OCTET STRING 
} 
 
UniformSendDataIndication  ::=  [APPLICATION 28] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 channelId    ChannelId, 
 dataPriority     DataPriority, 
 segmentation   Segmentation, 
 userData    OCTET STRING 
} 
 
--  Part 8: Token management 
 
TokenGrabRequest  ::=  [APPLICATION 29] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId 
} 
 
TokenGrabConfirm  ::=  [APPLICATION 30] IMPLICIT SEQUENCE 
{ 
 result  Result, 
 initiator    UserId, 
 tokenId    TokenId, 
 tokenStatus    TokenStatus 
} 

TokenInhibitRequest  ::=  [APPLICATION 31] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId 
} 
 
TokenInhibitConfirm  ::=  [APPLICATION 32] IMPLICIT SEQUENCE 
{ 
 result      Result, 
 initiator    UserId, 
 tokenId    TokenId, 
 tokenStatus    TokenStatus 
} 
 
TokenGiveRequest  ::=  [APPLICATION 33] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId, 
 recipient    UserId 
} 
 
TokenGiveIndication  ::=  [APPLICATION 34] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId, 
 recipient    UserId 
} 
 
TokenGiveResponse  ::=  [APPLICATION 35] IMPLICIT SEQUENCE 
{ 
 result      Result, 
 recipient    UserId, 
 tokenId    TokenId 
} 
 
TokenGiveConfirm  ::=  [APPLICATION 36] IMPLICIT SEQUENCE 
{ 
 result      Result, 
 initiator    UserId, 
 tokenId    TokenId, 
 tokenStatus    TokenStatus 
} 
 
TokenPleaseRequest  ::=  [APPLICATION 37] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId 
} 
 
TokenPleaseIndication  ::=  [APPLICATION 38] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId 
} 
 
TokenReleaseRequest  ::=  [APPLICATION 39] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId 
} 

TokenReleaseConfirm  ::=  [APPLICATION 40] IMPLICIT SEQUENCE 
{ 
 result  Result, 
 initiator    UserId, 
 tokenId    TokenId, 
 tokenStatus    TokenStatus 
} 
 
TokenTestRequest  ::=  [APPLICATION 41] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId 
} 
 
TokenTestConfirm  ::=  [APPLICATION 42] IMPLICIT SEQUENCE 
{ 
 initiator    UserId, 
 tokenId    TokenId, 
 tokenStatus    TokenStatus 
} 
 
--  Part 9: Status codes 
 
Reason  ::=  ENUMERATED  --  in DisconnectProviderUltimatum, DetachUserRequest, DetachUserIndication 
{ 
 rn-domain-disconnected  (0), 
 rn-provider-initiated  (1), 
 rn-token-purged   (2), 

 rn-user-requested   (3), 
 rn-channel-purged    (4) 
} 
 
Result  ::=  ENUMERATED  --  in Connect, response, confirm 
{ 
 rt-successful     (0), 
 rt-domain-merging    (1), 
 rt-domain-not-hierarchical (2), 
 rt-no-such-channel    (3), 
 rt-no-such-domain    (4), 
 rt-no-such-user   (5), 
 rt-not-admitted   (6), 
 rt-other-user-id   (7), 
 rt-parameters-unacceptable (8), 
 rt-token-not-available  (9), 
 rt-token-not-possessed  (10), 
 rt-too-many-channels  (11), 
 rt-too-many-tokens   (12), 
 rt-too-many-users   (13), 
 rt-unspecified-failure  (14), 
 rt-user-rejected   (15) 
} 
 
Diagnostic  ::=  ENUMERATED  --  in RejectMCSPDUUltimatum 
{ 
 dc-inconsistent-merge  (0), 
 dc-forbidden-PDU-downward (1), 
 dc-forbidden-PDU-upward (2), 
 dc-invalid-BER-encoding   (3), 
 dc-invalid-PER-encoding  (4), 
 dc-misrouted-user   (5), 
 dc-unrequested-confirm  (6), 
 dc-wrong-transport-priority (7), 
 dc-channel-id-conflict  (8), 
 dc-token-id-conflict   (9), 
 dc-not-user-id-channel  (10), 
 dc-too-many-channels  (11), 
 dc-too-many-tokens  (12), 
 dc-too-many-users   (13) 
} 
 
--  Part 10: MCSPDU repertoire 
 
ConnectMCSPDU  ::=  CHOICE 
{ 
 connect-initial   Connect-Initial, 
 connect-response   Connect-Response, 
 connect-additional    Connect-Additional, 
 connect-result   Connect-Result 
} 
 
DomainMCSPDU  ::=  CHOICE 
{ 
 plumbDomainIndication  PlumbDomainIndication, 
 erectDomainRequest  ErectDomainRequest, 
 mergeChannelsRequest  MergeChannelsRequest, 
 mergeChannelsConfirm  MergeChannelsConfirm, 
 purgeChannelsIndication   PurgeChannelsIndication, 

 mergeTokensRequest  MergeTokensRequest, 
 mergeTokensConfirm  MergeTokensConfirm, 
 purgeTokensIndication  PurgeTokensIndication, 
 disconnectProviderUltimatum DisconnectProviderUltimatum, 
 rejectMCSPDUUltimatum RejectMCSPDUUltimatum, 
 attachUserRequest    AttachUserRequest, 
 attachUserConfirm   AttachUserConfirm, 
 detachUserRequest   DetachUserRequest, 
 detachUserIndication  DetachUserIndication, 
 channelJoinRequest  ChannelJoinRequest, 
 channelJoinConfirm  ChannelJoinConfirm, 
 channelLeaveRequest  ChannelLeaveRequest, 
 channelConveneRequest  ChannelConveneRequest, 
 channelConveneConfirm  ChannelConveneConfirm, 
 channelDisbandRequest  ChannelDisbandRequest, 
 channelDisbandIndication  ChannelDisbandIndication, 
 channelAdmitRequest  ChannelAdmitRequest, 
 channelAdmitIndication  ChannelAdmitIndication, 
 channelExpelRequest  ChannelExpelRequest, 
 channelExpelIndication  ChannelExpelIndication, 
 sendDataRequest   SendDataRequest, 
 sendDataIndication   SendDataIndication, 
 uniformSendDataRequest  UniformSendDataRequest, 
 uniformSendDataIndication UniformSendDataIndication, 
 tokenGrabRequest    TokenGrabRequest, 
 tokenGrabConfirm   TokenGrabConfirm, 
 tokenInhibitRequest  TokenInhibitRequest, 
 tokenInhibitConfirm  TokenInhibitConfirm, 
 tokenGiveRequest   TokenGiveRequest, 
 tokenGiveIndication  TokenGiveIndication, 
 tokenGiveResponse   TokenGiveResponse, 
 tokenGiveConfirm    TokenGiveConfirm, 
 tokenPleaseRequest   TokenPleaseRequest, 
 tokenPleaseIndication  TokenPleaseIndication, 
 tokenReleaseRequest  TokenReleaseRequest, 
 tokenReleaseConfirm  TokenReleaseConfirm, 
 tokenTestRequest   TokenTestRequest, 
 tokenTestConfirm   TokenTestConfirm 
} 
 
END
"""


import asyncio
import traceback
from typing import cast, Dict, List

from aardwolf import logger
from aardwolf.channels import Channel
from aardwolf.protocol.T124.userdata.constants import ChannelOption
from aardwolf.extensions.RDPEDYC.protocol import DYNVC_MESSAGE, DYNVC_CMD
from aardwolf.extensions.RDPEDYC.protocol.caps import DYNVC_CAPS_REQ
from aardwolf.extensions.RDPEDYC.protocol.create import DYNVC_CREATE_REQ, DYNVC_CREATE_RSP
from aardwolf.extensions.RDPEDYC.protocol.close import DYNVC_CLOSE

from aardwolf.protocol.channelpdu import CHANNEL_PDU_HEADER, CHANNEL_FLAG
from aardwolf.protocol.T128.security import TS_SECURITY_HEADER,SEC_HDR_FLAG
from aardwolf.extensions.RDPEDYC.vchannels import VirtualChannelBase

class RDPEDYCChannel(Channel):
	name = 'drdynvc'
	def __init__(self, iosettings):
		Channel.__init__(self, self.name, ChannelOption.INITIALIZED|ChannelOption.ENCRYPT_RDP)
		self.defined_channels = iosettings.vchannels
		self.compression_needed = False
		self.version = 1
		self.version_data_sent = False
		self.channels:Dict[int, VirtualChannelBase] = {}
	
	async def start(self):
		try:
			#print('START called!')
			return True, None
		except Exception as e:
			return None, e

	async def stop(self):
		try:
			return True, None
		except Exception as e:
			return None, e

	async def close_channel(self, channelid):
		resp = DYNVC_CLOSE()
		resp.ChannelId = channelid
		await self.fragment_and_send(resp.to_bytes())
		if channelid not in self.channels:
			return
		await self.channels[channelid].channel_closed()

	async def send_channel_create_response(self, channelid, result:int = 0):
		resp = DYNVC_CREATE_RSP()
		resp.ChannelId = channelid
		resp.CreationStatus = result
		await self.fragment_and_send(resp.to_bytes())
	
	async def process_channel_data(self, data):
		#print('CHANNELDATA: %s' % data)
		channeldata = CHANNEL_PDU_HEADER.from_bytes(data)
		#print(channeldata)
		msg = DYNVC_MESSAGE.from_bytes(channeldata.data)
		#print(msg)
		if msg.cmd == DYNVC_CMD.CAPS_RSP and self.version_data_sent is False:
			await self.fragment_and_send(DYNVC_CAPS_REQ().to_bytes())
			self.version_data_sent = True
			return
		
		if msg.cmd == DYNVC_CMD.CREATE_RSP:
			msg = cast(DYNVC_CREATE_REQ, msg)
			if msg.ChannelName not in self.defined_channels:
				logger.debug('Server supports channel "%s" but client doesn\'t have definition!' % msg.ChannelName)
				await self.send_channel_create_response(msg.ChannelId, result = 0xC0000001) #STATUS_UNSUCCESSFUL
				return
			
			_, err = await self.defined_channels[msg.ChannelName].channel_init_internal(msg.ChannelId, self)
			if err is not None:
				logger.debug('Channel initialization failed! Error: %s' % err)
				await self.send_channel_create_response(msg.ChannelId, result = 0xC0000001) #STATUS_UNSUCCESSFUL
				return
				
			self.channels[msg.ChannelId] = self.defined_channels[msg.ChannelName]
			await self.send_channel_create_response(msg.ChannelId, result = 0) #STATUS_SUCCESS
			return

		elif msg.cmd == DYNVC_CMD.DATA_FIRST:
			if msg.ChannelId not in self.channels:
				logger.debug('DATA arrived for unknown channel %s' % msg.ChannelId)
				return
			await self.channels[msg.ChannelId].channel_rawdata_in(msg)
		elif msg.cmd == DYNVC_CMD.DATA:
			if msg.ChannelId not in self.channels:
				logger.debug('DATA arrived for unknown channel %s' % msg.ChannelId)
				return
			await self.channels[msg.ChannelId].channel_rawdata_in(msg)

		elif msg.cmd == DYNVC_CMD.CLOSE:
			if msg.ChannelId not in self.channels:
				logger.debug('CLOSE arrived for unknown channel %s' % msg.ChannelId)
				return
			await self.channels[msg.ChannelId].channel_closed()
		
		elif msg.cmd in [DYNVC_CMD.DATA_FIRST_COMPRESSED, DYNVC_CMD.DATA_COMPRESSED]:
			logger.debug('Compressed data recieved! Currently compression is not supported!')
		
		elif msg.cmd in [DYNVC_CMD.SOFT_SYNC_REQUEST, DYNVC_CMD.SOFT_SYNC_REQUEST]:
			logger.debug('Softsync packets are currently not supported')
		
		else:
			logger.debug('Unknown message type of %s recieved!' % msg.cmd)

	async def fragment_and_send(self, data):
		try:
			if len(data) < 16000:
				if self.compression_needed is False:
					flags = CHANNEL_FLAG.CHANNEL_FLAG_FIRST|CHANNEL_FLAG.CHANNEL_FLAG_LAST #|CHANNEL_FLAG.CHANNEL_FLAG_SHOW_PROTOCOL
					packet = CHANNEL_PDU_HEADER.serialize_packet(flags, data)
				else:
					raise NotImplementedError('Compression not implemented!')
			else:
				raise NotImplementedError('Chunked send not implemented!')
			
			sec_hdr = None
			if self.connection.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.send_channel_data(packet, sec_hdr, None, None, False)

			return True, False
		except Exception as e:
			traceback.print_exc()
			return None,e


import enum
from aardwolf.commons.queuedata import RDPDATATYPE
from aardwolf.commons.queuedata.constants import MOUSEBUTTON


class RDP_MOUSE:
	"""
	Mouse movement or button press event.
	xPos: mouse position X coordinate
	yPos: mouse position Y coordinate
	button: mouse button pressed/released. If only hovering use MOUSEBUTTON_LEFT and no press
	is_pressed: indicates wether the button has been pressed or released
	"""
	def __init__(self):
		self.type = RDPDATATYPE.MOUSE
		self.xPos:int = None
		self.yPos:int = True
		self.button:MOUSEBUTTON = None
		self.is_pressed:bool = None

	def __repr__(self):
		t = '==== RDP_MOUSE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
from typing import List
import enum
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, ORIENTATION

class TS_MONITOR_ATTRIBUTES:
	def __init__(self):
		self.physicalWidth:int = None
		self.physicalHeight:int = None
		self.orientation:ORIENTATION = None
		self.desktopScaleFactor:int = None
		self.deviceScaleFactor:int = None
		
	def to_bytes(self):
		t = self.physicalWidth.to_bytes(4, byteorder='little', signed = False)
		t += self.physicalHeight.to_bytes(4, byteorder='little', signed = False)
		t += self.orientation.value.to_bytes(4, byteorder='little', signed = False)
		t += self.desktopScaleFactor.to_bytes(4, byteorder='little', signed = False)
		t += self.deviceScaleFactor.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_MONITOR_ATTRIBUTES.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_MONITOR_ATTRIBUTES()
		msg.physicalWidth = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.physicalHeight = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.orientation = ORIENTATION(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.desktopScaleFactor = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.deviceScaleFactor = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_MONITOR_ATTRIBUTES ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/dfaf8842-c20c-4626-bd3b-8b7d0463bc0f
class TS_UD_CS_MONITOR_EX:
	def __init__(self):
		self.type:TS_UD_TYPE = TS_UD_TYPE.CS_MONITOR_EX
		self.length:int = None
		self.flags = 0 #not used
		self.monitorAttributeSize = None
		self.monitorCount:int = None
		self.monitorAttributesArray:List[TS_MONITOR_ATTRIBUTES] = []
		
	def to_bytes(self):
		def finish(t):
			t = (len(t)+4).to_bytes(2, byteorder='little', signed = False) + t
			t = self.type.value.to_bytes(2, byteorder='little', signed = False) + t
			return t
		t = self.flags.to_bytes(4, byteorder='little', signed = False)
		t += self.monitorAttributeSize.to_bytes(4, byteorder='little', signed = False)
		t += len(self.monitorCount).to_bytes(4, byteorder='little', signed = False)
		for cd in self.monitorAttributesArray:
			t += cd.to_bytes()
		return finish(t)

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UD_CS_MONITOR_EX.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UD_CS_MONITOR_EX()
		msg.type = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.flags = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.monitorAttributeSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.monitorCount = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		for _ in range(msg.monitorCount):
			msg.monitorAttributesArray.append(TS_MONITOR_ATTRIBUTES.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== TS_UD_CS_MONITOR_EX ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



from aardwolf.extensions.RDPEDYC.vchannels import VirtualChannelBase


class VchannelECHO(VirtualChannelBase):
	def __init__(self):
		VirtualChannelBase.__init__(self, 'ECHO')
	
	async def channel_init(self):
		#print('ECHO channel started!')
		return True, None

	async def channel_data_in(self, data):
		#print('ECHO channel data in: %s' % data)
		await self.channel_data_out(data)
	


import datetime

def timestamp2datetime(dt):
	"""
	Converting Windows timestamps to datetime.datetime format
	:param dt: Windows timestamp as array of bytes
	:type dt: bytearray
	:return: datetime.datetime
	"""
	us = int.from_bytes(dt, byteorder='little')/ 10.
	return datetime.datetime(1601, 1, 1) + datetime.timedelta(microseconds=us)
	
	
def datetime2timestamp(dt):
	delta = dt - datetime.datetime(1601, 1, 1)
	ns = int((delta / datetime.timedelta(microseconds=1)) * 10)
	return ns.to_bytes(8, 'little', signed = False)



import io
import copy
import typing
import asyncio
import traceback
from typing import cast
from collections import OrderedDict

import asn1tools
from PIL import Image
from aardwolf import logger
from aardwolf.keyboard import VK_MODIFIERS
from aardwolf.commons.queuedata.constants import MOUSEBUTTON, VIDEO_FORMAT
from aardwolf.commons.target import RDPTarget
from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret, asyauthProtocol
from aardwolf.commons.cryptolayer import RDPCryptoLayer
from aardwolf.network.x224 import X224Network
from asyauth.common.credentials.credssp import CREDSSPCredential

from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS, NEG_FLAGS
from aardwolf.protocol.x224.server.connectionconfirm import RDP_NEG_RSP

from aardwolf.protocol.pdu.input.keyboard import TS_KEYBOARD_EVENT, KBDFLAGS
from aardwolf.protocol.pdu.input.unicode import TS_UNICODE_KEYBOARD_EVENT
from aardwolf.protocol.pdu.input.mouse import PTRFLAGS, TS_POINTER_EVENT
from aardwolf.protocol.pdu.capabilities import CAPSTYPE
from aardwolf.protocol.pdu.capabilities.general import TS_GENERAL_CAPABILITYSET, OSMAJORTYPE, OSMINORTYPE, EXTRAFLAG
from aardwolf.protocol.pdu.capabilities.bitmap import TS_BITMAP_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.sound import TS_SOUND_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.virtualchannel import TS_VIRTUALCHANNEL_CAPABILITYSET, VCCAPS
from aardwolf.protocol.pdu.capabilities.offscreen import TS_OFFSCREEN_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.glyph import TS_GLYPHCACHE_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.brush import TS_BRUSH_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.input import TS_INPUT_CAPABILITYSET, INPUT_FLAG
from aardwolf.protocol.pdu.capabilities.pointer import TS_POINTER_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.bitmapcache import TS_BITMAPCACHE_CAPABILITYSET
from aardwolf.protocol.pdu.capabilities.order import TS_ORDER_CAPABILITYSET, ORDERFLAG

from aardwolf.protocol.T124.GCCPDU import GCCPDU
from aardwolf.protocol.T124.userdata import TS_UD, TS_SC
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE, HIGH_COLOR_DEPTH, ENCRYPTION_FLAG, SUPPORTED_COLOR_DEPTH, \
	COLOR_DEPTH, CONNECTION_TYPE, RNS_UD_CS, ClusterInfo
from aardwolf.protocol.T124.userdata.clientcoredata import TS_UD_CS_CORE
from aardwolf.protocol.T124.userdata.clientsecuritydata import TS_UD_CS_SEC
from aardwolf.protocol.T124.userdata.clientnetworkdata import TS_UD_CS_NET, CHANNEL_DEF
from aardwolf.protocol.T124.userdata.clientclusterdata import TS_UD_CS_CLUSTER
from aardwolf.protocol.T128.security import TS_SECURITY_HEADER,SEC_HDR_FLAG, TS_SECURITY_HEADER1
from aardwolf.protocol.T125.infopacket import TS_INFO_PACKET, INFO_FLAG
from aardwolf.protocol.T125.extendedinfopacket import TS_EXTENDED_INFO_PACKET, TS_TIME_ZONE_INFORMATION, TS_SYSTEMTIME, CLI_AF
from aardwolf.protocol.T125.MCSPDU_ver_2 import MCSPDU_ver_2
from aardwolf.protocol.T128.serverdemandactivepdu import TS_DEMAND_ACTIVE_PDU
from aardwolf.protocol.T128.clientconfirmactivepdu import TS_SHARECONTROLHEADER, TS_CONFIRM_ACTIVE_PDU, TS_CAPS_SET
from aardwolf.protocol.T128.synchronizepdu import TS_SYNCHRONIZE_PDU
from aardwolf.protocol.T128.controlpdu import TS_CONTROL_PDU, CTRLACTION
from aardwolf.protocol.T128.fontlistpdu import TS_FONT_LIST_PDU
from aardwolf.protocol.T128.inputeventpdu import TS_SHAREDATAHEADER, TS_INPUT_EVENT, TS_INPUT_PDU_DATA
from aardwolf.protocol.T125.securityexchangepdu import TS_SECURITY_PACKET
from aardwolf.protocol.T128.seterrorinfopdu import TS_SET_ERROR_INFO_PDU
from aardwolf.protocol.T128.share import PDUTYPE, STREAM_TYPE, PDUTYPE2



from aardwolf.protocol.fastpath import TS_FP_UPDATE_PDU, FASTPATH_UPDATETYPE, FASTPATH_FRAGMENT, FASTPATH_SEC, TS_FP_UPDATE
from aardwolf.commons.queuedata import RDPDATATYPE, RDP_KEYBOARD_SCANCODE, RDP_KEYBOARD_UNICODE, RDP_MOUSE, RDP_VIDEO
from aardwolf.channels import MCSChannel
from aardwolf.commons.iosettings import RDPIOSettings

from asysocks.unicomm.client import UniClient
from asysocks.unicomm.common.connection import UniConnection
from aardwolf.network.tpkt import TPKTPacketizer

from aardwolf.network.tpkt import CredSSPPacketizer
from asysocks.unicomm.common.packetizers import Packetizer

class RDPConnection:
	def __init__(self, target:RDPTarget, credentials:UniCredential, iosettings:RDPIOSettings):
		"""RDP client connection object. After successful connection the two asynchronous queues named `ext_out_queue` and `ext_in_queue`
		can be used to communicate with the remote server

		Args:
			target (RDPTarget): Target object specifying the network connection details
			credentials (RDPCredential): Credential object specifying the authentication details
			iosettings (RDPIOSettings): Screen/Keyboard/IO settings
		"""
		self.target = target
		self.credentials = credentials
		self.authapi = None
		self.iosettings = iosettings
		self.disconnected_evt = asyncio.Event() #this will be set if we disconnect for whatever reason

		# these are the main queues with which you can communicate with the server
		# ext_out_queue: yields video,clipboard,bell data
		# ext_in_queue: expects keyboard/mouse/clipboard data
		self.ext_out_queue = asyncio.Queue()
		self.ext_in_queue = asyncio.Queue()

		self.__connection:UniConnection = None
		self._x224net = None
		self.__t125_ber_codec = None
		self._t125_per_codec = None
		self.__t124_codec = None

		self.x224_connection_reply = None
		self.x224_protocol = None

		self.__server_connect_pdu:TS_SC = None # serverconnectpdu message from server (holds security exchange data)
		
		self._initiator = None
		self.__channel_id_lookup = {}
		self.__joined_channels =  OrderedDict({})
		
		for channel in self.iosettings.channels:
			self.__joined_channels[channel.name] = channel(self.iosettings)
		
		self.__channel_task = {} #name -> channeltask

		
		self.__external_reader_task = None
		self.__x224_reader_task = None
		self.client_x224_flags = 0
		self.client_x224_supported_protocols = self.iosettings.supported_protocols 
		self.cryptolayer:RDPCryptoLayer = None
		self.__desktop_buffer = None
		self.desktop_buffer_has_data = False

		self.__vk_to_sc = {
			'VK_BACK'     : 14,
			'VK_ESCAPE'   : 1,
			'VK_TAB'      : 15,
			'VK_RETURN'   : 28,
			'VK_INSERT'   : 82,
			'VK_DELETE'   : 83,
			'VK_HOME'     : 71,
			'VK_END'      : 79,
			'VK_PRIOR'    : 73,
			'VK_NEXT'     : 81,
			'VK_LEFT'     : 75,
			'VK_UP'       : 72,
			'VK_RIGHT'    : 77,
			'VK_DOWN'     : 80,
			'VK_F1'       : 59,
			'VK_F2'       : 60,
			'VK_F3'       : 61,
			'VK_F4'       : 62,
			'VK_F5'       : 63,
			'VK_F6'       : 64,
			'VK_F7'       : 65,
			'VK_F8'       : 66,
			'VK_F9'       : 67,
			'VK_F10'      : 68,
			'VK_F11'      : 87,
			'VK_F12'      : 88,
			'VK_LSHIFT'   : 42,
			'VK_RSHIFT'   : 54,
			'VK_LCONTROL' : 29,
			'VK_LWIN'     : 57435,
			'VK_RWIN'     : 57436,
			'VK_LMENU'    : 56,
			'VK_SCROLL'   : 70,
			'VK_NUMLOCK'  : 69,
			'VK_CAPITAL'  : 58,
			'VK_RCONTROL' : 57629,
			'VK_MULTIPLY' : 55,
			'VK_ADD'      : 78,
			'VK_SUBTRACT' : 74,
			'VK_DIVIDE'   : 57397,
			'VK_SNAPSHOT' : 84,
			#'VK_RCONTROL' : 57373,
			#'VK_PAUSE'    : 57629,
			'VK_RMENU'    : 57400,
			#'VK_DBE_NOCODEINPUT': 98, # except on KLID: 00000412 (ko)
			#'VK_DECIMAL' not found anywhere?
		}

	
	async def terminate(self):
		try:
			for name in self.__joined_channels:
				await self.__joined_channels[name].disconnect()
			
			if self.ext_out_queue is not None:
				# signaling termination via ext_out_queue
				await self.ext_out_queue.put(None)
				
			
			if self.__external_reader_task is not None:
				self.__external_reader_task.cancel()
			
			if self.__x224_reader_task is not None:
				self.__x224_reader_task.cancel()
			
			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e
		finally:
			self.disconnected_evt.set()
	
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await asyncio.wait_for(self.terminate(), timeout = 5)
	
	async def connect(self):
		"""Initiates the connection to the server, and performs authentication and all necessary setups.
		Returns:
			Tuple[bool, Exception]: _description_
		"""
		try:

			packetizer = TPKTPacketizer()
			client = UniClient(self.target, packetizer)
			self.__connection = await client.connect()

			# X224 channel is on top of TPKT, performs the initial negotiation
			# between the server and our client (restricted admin mode, authentication methods etc)
			# are set here
			self._x224net = X224Network(self.__connection)
			if self.client_x224_supported_protocols is None and self.credentials is not None:
				if self.credentials.protocol in [asyauthProtocol.NTLM, asyauthProtocol.KERBEROS]:
					if self.credentials.secret is not None and self.credentials.stype not in [asyauthSecret.PASSWORD, asyauthSecret.PWPROMPT, asyauthSecret.PWHEX, asyauthSecret.PWB64]:
						# user provided some secret but it's not a password
						# here we request restricted admin mode
						self.client_x224_flags = NEG_FLAGS.RESTRICTED_ADMIN_MODE_REQUIRED
						self.client_x224_supported_protocols = SUPP_PROTOCOLS.RDP | SUPP_PROTOCOLS.SSL |SUPP_PROTOCOLS.HYBRID
					else:
						self.client_x224_flags = 0
						self.client_x224_supported_protocols = SUPP_PROTOCOLS.RDP | SUPP_PROTOCOLS.SSL | SUPP_PROTOCOLS.HYBRID_EX | SUPP_PROTOCOLS.HYBRID
				
				elif self.credentials.stype == asyauthSecret.NONE: #and self.credentials.username is None:
					# not sending any passwords, hoping HYBRID is not required
					self.client_x224_flags = 0
					self.client_x224_supported_protocols = SUPP_PROTOCOLS.RDP | SUPP_PROTOCOLS.SSL
				else:
					self.client_x224_flags = 0
					self.client_x224_supported_protocols = SUPP_PROTOCOLS.RDP | SUPP_PROTOCOLS.SSL
			
			logger.debug('Client protocol flags: %s' % self.client_x224_flags)
			logger.debug('Client protocol offer: %s' % self.client_x224_supported_protocols)
			connection_accepted_reply, err = await self._x224net.client_negotiate(self.client_x224_flags, self.client_x224_supported_protocols)
			if err is not None:
				raise err
			
			if connection_accepted_reply.rdpNegData is not None:
				# newer RDP protocol was selected

				self.x224_connection_reply = typing.cast(RDP_NEG_RSP, connection_accepted_reply.rdpNegData)
				# if the server requires SSL/TLS connection as indicated in the 'selectedProtocol' flags
				# we switch here. SSL and HYBRID/HYBRID_EX authentication methods all require this switch
				
				
				self.x224_protocol = self.x224_connection_reply.selectedProtocol
				self.x224_flag = self.x224_connection_reply.flags
				logger.debug('Server selected protocol: %s' % self.x224_protocol)
				if SUPP_PROTOCOLS.SSL in self.x224_protocol or SUPP_PROTOCOLS.HYBRID in self.x224_protocol or SUPP_PROTOCOLS.HYBRID_EX in self.x224_protocol:
					await self.__connection.wrap_ssl()

				# if the server expects HYBRID/HYBRID_EX authentication we do that here
				# This is basically credSSP
				if SUPP_PROTOCOLS.HYBRID in self.x224_protocol or SUPP_PROTOCOLS.HYBRID_EX in self.x224_protocol:
					_, err = await self.credssp_auth()
					if err is not None:
						raise err
					
					#switching back to tpkt
					self.__connection.change_packetizer(TPKTPacketizer())

			else:
				# old RDP protocol is used
				self.x224_protocol = SUPP_PROTOCOLS.RDP
				self.x224_flag = None

			# initializing the parsers here otherwise they'd waste time on connections that did not get to this point
			# not kidding, this takes ages
			self.__t125_ber_codec = asn1tools.compile_string(MCSPDU_ver_2, 'ber')
			self._t125_per_codec = asn1tools.compile_string(MCSPDU_ver_2, 'per')
			self.__t124_codec = asn1tools.compile_string(GCCPDU, 'per')

			# All steps below are required as stated in the following 'documentation'
			# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/1d263f84-6153-4a16-b329-8770be364e1b
			logger.debug('Establish channels...')
			_, err = await self.__establish_channels()
			if err is not None:
				raise err
			logger.debug('Establish channels OK')
			
			_, err = await self.__erect_domain()
			if err is not None:
				raise err
			logger.debug('Erect domain OK')
			
			_, err = await self.__attach_user()
			if err is not None:
				raise err
			logger.debug('Attach user OK')
			
			_, err = await self.__join_channels()
			if err is not None:
				raise err
			logger.debug('Join channels OK')
			
			if self.x224_protocol == SUPP_PROTOCOLS.RDP:
				# key exchange here because we use old version of the protocol
				_, err = await self.__security_exchange()
				if err is not None:
					raise err
				logger.debug('Security exchange OK')

			_, err = await self.__send_userdata()
			if err is not None:
				raise err
			logger.debug('Send userdata OK')

			_, err = await self.__handle_license()
			if err is not None:
				raise err
			logger.debug('handle license OK')

			_, err = await self.__handle_mandatory_capability_exchange()
			if err is not None:
				raise err
			logger.debug('mandatory capability exchange OK')

			self.__external_reader_task = asyncio.create_task(self.__external_reader())
			logger.debug('RDP connection sequence done')
			self.__desktop_buffer = Image.new(mode="RGBA", size=(self.iosettings.video_width, self.iosettings.video_height))
			return True, None
		except Exception as e:
			self.disconnected_evt.set()
			return None, e
	
	def get_extra_info(self):
		# for NTLM fingerprinting
		ntlm_data = self.authapi.get_extra_info()
		if ntlm_data is not None:
			return ntlm_data.to_dict()
		return None
		 
	
	async def credssp_auth(self):
		try:
			#constructing authentication API is not specified
			if self.authapi is None:
				if self.credentials is None:
					raise Exception('No auth API nor credentials were supplied!')
				
				
				self.authapi = CREDSSPCredential([self.credentials]).build_context()

			self.__connection.change_packetizer(CredSSPPacketizer())

			certificate = self.__connection.get_peer_certificate()

			# credSSP auth happends here
			token = None
			data, to_continue, err = await self.authapi.authenticate(token, flags = None, certificate = certificate, spn=self.target.to_target_string())
			if err is not None:
				raise err

			await self.__connection.write(data)
			
			for _ in range(10):
				token = await self.__connection.read_one()
				data, to_continue, err = await self.authapi.authenticate(token, flags = None, certificate = certificate, spn=self.target.to_target_string())
				if err is not None:
					raise err
				
				if to_continue is False:
					# credSSP auth finished, flushing remaining data
					if data is not None:
						await self.__connection.write(data)
					
					# if HYBRID_EX auth was selected by the server, the server MUST send
					# an extra packet informing us if the credSSP auth was successful or not
					if SUPP_PROTOCOLS.HYBRID_EX in self.x224_protocol:
						self.__connection.change_packetizer(Packetizer())
						authresult_raw = await self.__connection.read_one()
						authresult = int.from_bytes(authresult_raw, byteorder='little', signed=False)
						#print('Early User Authorization Result PDU %s' % authresult)
						if authresult == 5:
							raise Exception('Authentication failed! (early user auth)')
					return True, None
				
				await self.__connection.write(data)

		except Exception as e:
			return None, e

	async def __establish_channels(self):
		try:
			ts_ud = TS_UD()

			ud_core = TS_UD_CS_CORE()
			ud_core.desktopWidth = self.iosettings.video_width
			ud_core.desktopHeight = self.iosettings.video_height
			# this part doesn matter since we also set postBeta2ColorDepth
			#ud_core.colorDepth = COLOR_DEPTH.COLOR_8BPP
			if self.iosettings.video_bpp_min == 4:
				ud_core.colorDepth = COLOR_DEPTH.COLOR_4BPP
			elif self.iosettings.video_bpp_min == 8:
				ud_core.colorDepth = COLOR_DEPTH.COLOR_8BPP
			elif self.iosettings.video_bpp_min == 15:
				ud_core.colorDepth = COLOR_DEPTH.COLOR_16BPP_555
			elif self.iosettings.video_bpp_min == 16:
				ud_core.colorDepth = COLOR_DEPTH.COLOR_16BPP_565
			elif self.iosettings.video_bpp_min == 24:
				ud_core.colorDepth = COLOR_DEPTH.COLOR_24BPP
			# from here on it matters

			ud_core.keyboardLayout = self.iosettings.keyboard_layout
			ud_core.clientBuild = 2600
			ud_core.clientName = 'aardworlf'
			ud_core.imeFileName = ''
			#ud_core.postBeta2ColorDepth = COLOR_DEPTH.COLOR_8BPP
			if self.iosettings.video_bpp_min == 4:
				ud_core.postBeta2ColorDepth = COLOR_DEPTH.COLOR_4BPP
			elif self.iosettings.video_bpp_min == 8:
				ud_core.postBeta2ColorDepth = COLOR_DEPTH.COLOR_8BPP
			elif self.iosettings.video_bpp_min == 15:
				ud_core.postBeta2ColorDepth = COLOR_DEPTH.COLOR_16BPP_555
			elif self.iosettings.video_bpp_min == 16:
				ud_core.postBeta2ColorDepth = COLOR_DEPTH.COLOR_16BPP_565
			elif self.iosettings.video_bpp_min == 24:
				ud_core.postBeta2ColorDepth = COLOR_DEPTH.COLOR_24BPP

			ud_core.clientProductId = 1
			ud_core.serialNumber = 0
			ud_core.highColorDepth = HIGH_COLOR_DEPTH.HIGH_COLOR_16BPP

			if self.iosettings.video_bpp_max == 4:
				ud_core.highColorDepth = HIGH_COLOR_DEPTH.HIGH_COLOR_4BPP
			elif self.iosettings.video_bpp_max == 8:
				ud_core.highColorDepth = HIGH_COLOR_DEPTH.HIGH_COLOR_8BPP
			elif self.iosettings.video_bpp_max == 15:
				ud_core.highColorDepth = HIGH_COLOR_DEPTH.HIGH_COLOR_15BPP
			elif self.iosettings.video_bpp_max == 16:
				ud_core.highColorDepth = HIGH_COLOR_DEPTH.HIGH_COLOR_16BPP
			elif self.iosettings.video_bpp_max == 24:
				ud_core.highColorDepth = HIGH_COLOR_DEPTH.HIGH_COLOR_24BPP

			self.iosettings.video_bpp_supported.append(self.iosettings.video_bpp_max)
			self.iosettings.video_bpp_supported.append(self.iosettings.video_bpp_min)
			ud_core.supportedColorDepths = SUPPORTED_COLOR_DEPTH.RNS_UD_16BPP_SUPPORT
			for sc in self.iosettings.video_bpp_supported:
				if sc == 15:
					ud_core.supportedColorDepths |= SUPPORTED_COLOR_DEPTH.RNS_UD_15BPP_SUPPORT
				elif sc == 16:
					ud_core.supportedColorDepths |= SUPPORTED_COLOR_DEPTH.RNS_UD_16BPP_SUPPORT
				elif sc == 24:
					ud_core.supportedColorDepths |= SUPPORTED_COLOR_DEPTH.RNS_UD_24BPP_SUPPORT
				elif sc == 32:
					ud_core.supportedColorDepths |= SUPPORTED_COLOR_DEPTH.RNS_UD_32BPP_SUPPORT
			
			ud_core.earlyCapabilityFlags = RNS_UD_CS.SUPPORT_ERRINFO_PDU
			ud_core.clientDigProductId = b'\x00' * 64
			ud_core.connectionType = CONNECTION_TYPE.UNK
			ud_core.pad1octet = b'\x00'
			ud_core.serverSelectedProtocol = self.x224_protocol
			
			ud_sec = TS_UD_CS_SEC()
			ud_sec.encryptionMethods = ENCRYPTION_FLAG.FRENCH if self.x224_protocol is not SUPP_PROTOCOLS.RDP else ENCRYPTION_FLAG.BIT_128
			ud_sec.extEncryptionMethods = ENCRYPTION_FLAG.FRENCH

			ud_clust = TS_UD_CS_CLUSTER()
			ud_clust.RedirectedSessionID = 0
			ud_clust.Flags = 8|4|ClusterInfo.REDIRECTION_SUPPORTED

			ud_net = TS_UD_CS_NET()
			
			for name in self.__joined_channels:
				cd = CHANNEL_DEF()
				cd.name = name
				cd.options = self.__joined_channels[name].options
				ud_net.channelDefArray.append(cd)
			

			ts_ud.userdata = {
				TS_UD_TYPE.CS_CORE : ud_core,
				TS_UD_TYPE.CS_SECURITY : ud_sec,
				TS_UD_TYPE.CS_CLUSTER : ud_clust,
				TS_UD_TYPE.CS_NET : ud_net
			}

			userdata_wrapped = {
				'conferenceName': {
					'numeric': '0'
				}, 
				'lockedConference': False, 
				'listedConference': False, 
				'conductibleConference': False, 
				'terminationMethod': 'automatic', 
				'userData': [
					{
						'key': ('h221NonStandard', b'Duca'), 
						'value': ts_ud.to_bytes()
					}
				]
			}

			ConnectGCCPDU = self.__t124_codec.encode('ConnectGCCPDU', ('conferenceCreateRequest', userdata_wrapped))
			t124_wrapper = {
				't124Identifier': ('object', '0.0.20.124.0.1'), 
				'connectPDU': ConnectGCCPDU
			}
			t124_wrapper = self.__t124_codec.encode('ConnectData', t124_wrapper)

			initialconnect = {
				'callingDomainSelector': b'\x01', 
				'calledDomainSelector': b'\x01', 
				'upwardFlag': True, 
				'targetParameters': {
					'maxChannelIds': 34, 
					'maxUserIds': 2, 
					'maxTokenIds': 0, 
					'numPriorities': 1, 
					'minThroughput': 0, 
					'maxHeight': 1, 
					'maxMCSPDUsize': -1, 
					'protocolVersion': 2
				}, 
				'minimumParameters': {
					'maxChannelIds': 1, 
					'maxUserIds': 1, 
					'maxTokenIds': 1, 
					'numPriorities': 1, 
					'minThroughput': 0, 
					'maxHeight': 1, 
					'maxMCSPDUsize': 1056, 
					'protocolVersion': 2
				}, 
				'maximumParameters': {
					'maxChannelIds': -1, 
					'maxUserIds': -1001, 
					'maxTokenIds': -1, 
					'numPriorities': 1, 
					'minThroughput': 0, 
					'maxHeight': 1, 
					'maxMCSPDUsize': -1, 
					'protocolVersion': 2
				}, 
				'userData': t124_wrapper
			}

			conf_create_req = self.__t125_ber_codec.encode('ConnectMCSPDU',('connect-initial', initialconnect))
			await self._x224net.write(bytes(conf_create_req))
			
			response_raw = await self._x224net.read()			
			server_res_raw = response_raw.data
			server_res_t125 = self.__t125_ber_codec.decode('ConnectMCSPDU', server_res_raw)
			if server_res_t125[0] != 'connect-response':
				raise Exception('Unexpected response! %s' % server_res_t125)
			if server_res_t125[1]['result'] != 'rt-successful':
				raise Exception('Server returned error! %s' % server_res_t125)
			
			server_res_t124 = self.__t124_codec.decode('ConnectData', server_res_t125[1]['userData'])
			if server_res_t124['t124Identifier'][1] != '0.0.20.124.0.1':
				raise Exception('Unexpected T124 response: %s' % server_res_t124)
			
			# this is strange, and it seems wireshark struggles here as well. 
			# it seems the encoding used does not account for all the packet 
			# bytes at the end but those are also needed for decoding the sub-strucutres?!

			data = server_res_t124['connectPDU']
			m = server_res_raw.find(data)
			remdata = server_res_raw[m+len(data):]
			
			# weirdness ends here... FOR NOW!

			server_connect_pdu_raw = self.__t124_codec.decode('ConnectGCCPDU', server_res_t124['connectPDU']+remdata)
			self.__server_connect_pdu = TS_SC.from_bytes(server_connect_pdu_raw[1]['userData'][0]['value']).serverdata
			
			logger.log(1, 'Server capability set: %s' % self.__server_connect_pdu)

			# populating channels
			scnet = self.__server_connect_pdu[TS_UD_TYPE.SC_NET]
			for i, name in enumerate(self.__joined_channels):
				self.__joined_channels[name].channel_id = scnet.channelIdArray[i]
				self.__channel_id_lookup[scnet.channelIdArray[i]] = self.__joined_channels[name]

			self.__joined_channels['MCS'] = MCSChannel() #TODO: options?
			self.__joined_channels['MCS'].channel_id = scnet.MCSChannelId
			self.__channel_id_lookup[scnet.MCSChannelId] = self.__joined_channels['MCS']

			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def __erect_domain(self):
		try:
			# the parser could not decode nor encode this data correctly.
			# therefore we are sending these as bytes. it's static 
			# (even according to docu)
			await self._x224net.write(bytes.fromhex('0400010001'))
			return True, None
		except Exception as e:
			return None, e
	
	async def __attach_user(self):
		try:
			request = self._t125_per_codec.encode('DomainMCSPDU', ('attachUserRequest', {}))
			await self._x224net.write(request)
			response = await self._x224net.read()
			response_parsed = self._t125_per_codec.decode('DomainMCSPDU', response.data)
			if response_parsed[0] != 'attachUserConfirm':
				raise Exception('Unexpected response! %s' % response_parsed[0])
			if response_parsed[1]['result'] != 'rt-successful':
				raise Exception('Server returned error! %s' % response_parsed[0])
			self._initiator = response_parsed[1]['initiator']
			
			return True, None
		except Exception as e:
			return None, e
	
	async def __join_channels(self):
		try:
			for name in self.__joined_channels:
				joindata = self._t125_per_codec.encode('DomainMCSPDU', ('channelJoinRequest', {'initiator': self._initiator, 'channelId': self.__joined_channels[name].channel_id}))
				await self._x224net.write(bytes(joindata))
				response = await self._x224net.read()
				
				x = self._t125_per_codec.decode('DomainMCSPDU', response.data)
				if x[0] != 'channelJoinConfirm':
					raise Exception('Could not join channel "%s". Reason: %s' % (name, x))
				
				self.__channel_task[name] = asyncio.create_task(self.__joined_channels[name].run(self))
				
			
			self.__x224_reader_task = asyncio.create_task(self.__x224_reader())
			return True, None
		except Exception as e:
			return None, e
	
	async def __security_exchange(self):
		try:
			self.cryptolayer = RDPCryptoLayer(self.__server_connect_pdu[TS_UD_TYPE.SC_SECURITY].serverRandom)
			enc_secret = self.__server_connect_pdu[TS_UD_TYPE.SC_SECURITY].serverCertificate.encrypt(self.cryptolayer.ClientRandom)
			secexchange = TS_SECURITY_PACKET()
			secexchange.encryptedClientRandom = enc_secret

			sec_hdr = TS_SECURITY_HEADER()
			sec_hdr.flags = SEC_HDR_FLAG.EXCHANGE_PKT
			sec_hdr.flagsHi = 0

			await self.handle_out_data(secexchange, sec_hdr, None, None, self.__joined_channels['MCS'].channel_id, False)
			return True, None
		except Exception as e:
			return None, e

	async def __send_userdata(self):
		try:
			systime = TS_SYSTEMTIME()
			systime.wYear = 0
			systime.wMonth = 10
			systime.wDayOfWeek = 0
			systime.wDay = 5
			systime.wHour = 3
			systime.wMinute = 0
			systime.wSecond = 0
			systime.wMilliseconds = 0

			systz = TS_TIME_ZONE_INFORMATION()
			systz.Bias = 4294967236
			systz.StandardName = b'G\x00T\x00B\x00,\x00 \x00s\x00o\x00m\x00m\x00a\x00r\x00t\x00i\x00d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
			systz.StandardDate = systime
			systz.StandardBias = 0
			systz.DaylightName = b'G\x00T\x00B\x00,\x00 \x00s\x00o\x00m\x00m\x00a\x00r\x00t\x00i\x00d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
			systz.DaylightDate = systime
			systz.DaylightBias = 4294967236

			extinfo = TS_EXTENDED_INFO_PACKET()
			extinfo.clientAddressFamily = CLI_AF.AF_INET
			extinfo.clientAddress = '10.10.10.101'
			extinfo.clientDir = 'C:\\WINNT\\System32\\mstscax.dll'
			extinfo.clientTimeZone = systz
			extinfo.clientSessionId = 0
			if self.iosettings.performance_flags is not None:
				extinfo.performanceFlags = self.iosettings.performance_flags

			info = TS_INFO_PACKET()
			info.CodePage = 0
			info.flags = INFO_FLAG.ENABLEWINDOWSKEY|INFO_FLAG.MAXIMIZESHELL|INFO_FLAG.UNICODE|INFO_FLAG.DISABLECTRLALTDEL|INFO_FLAG.MOUSE
			info.Domain = ''
			info.UserName = ''
			info.Password = ''
			if self.authapi is None or SUPP_PROTOCOLS.SSL in self.x224_protocol:
				if self.credentials.domain is not None:
					info.Domain = self.credentials.domain
				if self.credentials.username is not None:
					info.UserName = self.credentials.username
				if self.credentials.secret is not None:
					info.Password = self.credentials.secret
			info.AlternateShell = '' 
			info.WorkingDir = ''
			info.extrainfo = extinfo

			sec_hdr = TS_SECURITY_HEADER()
			sec_hdr.flags = SEC_HDR_FLAG.INFO_PKT
			if self.cryptolayer is not None:
				sec_hdr.flags |= SEC_HDR_FLAG.ENCRYPT
			sec_hdr.flagsHi = 0

			await self.handle_out_data(info, sec_hdr, None, None, self.__joined_channels['MCS'].channel_id, False)
			return True, None
		except Exception as e:
			return None, e

	async def __handle_license(self):
		try:
			# TODO: implement properly
			# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/7d941d0d-d482-41c5-b728-538faa3efb31
			data, err = await self.__joined_channels['MCS'].out_queue.get()
			if err is not None:
				raise err
			
			res = self._t125_per_codec.decode('DomainMCSPDU', data)
			if res[0] == 'tokenInhibitConfirm':
				if res[1]['result'] != 'rt-successful':
					raise Exception('License error! tokenInhibitConfirm:result not successful')
			else:
				raise Exception('tokenInhibitConfirm did not show up in reply!')

			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e
	
	async def __handle_mandatory_capability_exchange(self):
		try:
			# waiting for server to demand active pdu and inside send its capabilities
			data, err = await self.__joined_channels['MCS'].out_queue.get()
			if err is not None:
				raise err

			data_start_offset = 0
			if self.__server_connect_pdu[TS_UD_TYPE.SC_SECURITY].encryptionLevel == 1:
				# encryptionLevel == 1 means that server data is not encrypted. This results in this part of the negotiation 
				# that the server sends data to the client with an empty security header (which is not documented....)
				data_start_offset = 4

			data = data[data_start_offset:]
			shc = TS_SHARECONTROLHEADER.from_bytes(data)
			if shc.pduType != PDUTYPE.DEMANDACTIVEPDU:
				raise Exception('Unexpected reply! Expected DEMANDACTIVEPDU got "%s" instead!' % shc.pduType.name)
			
			res = TS_DEMAND_ACTIVE_PDU.from_bytes(data)
			for cap in res.capabilitySets:
				if cap.capabilitySetType == CAPSTYPE.GENERAL:
					cap = typing.cast(TS_GENERAL_CAPABILITYSET, cap.capability)
					if EXTRAFLAG.ENC_SALTED_CHECKSUM in cap.extraFlags and self.cryptolayer is not None:
						self.cryptolayer.use_encrypted_mac = True
			
			caps = []
			# now we send our capabilities
			cap = TS_GENERAL_CAPABILITYSET()
			cap.osMajorType = OSMAJORTYPE.WINDOWS
			cap.osMinorType = OSMINORTYPE.WINDOWS_NT
			cap.extraFlags =  EXTRAFLAG.FASTPATH_OUTPUT_SUPPORTED | EXTRAFLAG.NO_BITMAP_COMPRESSION_HDR | EXTRAFLAG.LONG_CREDENTIALS_SUPPORTED
			if self.cryptolayer is not None and self.cryptolayer.use_encrypted_mac is True:
				cap.extraFlags |= EXTRAFLAG.ENC_SALTED_CHECKSUM
			caps.append(cap)

			cap = TS_BITMAP_CAPABILITYSET()
			cap.preferredBitsPerPixel = self.iosettings.video_bpp_max
			cap.desktopWidth = self.iosettings.video_width
			cap.desktopHeight = self.iosettings.video_height
			caps.append(cap)

			#TS_FONT_CAPABILITYSET missing

			cap = TS_ORDER_CAPABILITYSET()
			cap.orderFlags = ORDERFLAG.ZEROBOUNDSDELTASSUPPORT | ORDERFLAG.NEGOTIATEORDERSUPPORT #do not change this!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			#TEST
			cap.orderFlags |= ORDERFLAG.SOLIDPATTERNBRUSHONLY
			caps.append(cap)

			cap = TS_BITMAPCACHE_CAPABILITYSET()
			caps.append(cap)

			cap = TS_POINTER_CAPABILITYSET()
			caps.append(cap)

			cap = TS_INPUT_CAPABILITYSET()
			cap.inputFlags = INPUT_FLAG.SCANCODES
			cap.keyboardLayout = self.iosettings.keyboard_layout
			cap.keyboardType = self.iosettings.keyboard_type
			cap.keyboardSubType = self.iosettings.keyboard_subtype
			cap.keyboardFunctionKey = self.iosettings.keyboard_functionkey
			caps.append(cap)

			cap = TS_BRUSH_CAPABILITYSET()
			caps.append(cap)

			cap = TS_GLYPHCACHE_CAPABILITYSET()
			caps.append(cap)

			cap = TS_OFFSCREEN_CAPABILITYSET()
			caps.append(cap)

			cap = TS_VIRTUALCHANNEL_CAPABILITYSET()
			cap.flags = VCCAPS.COMPR_CS_8K | VCCAPS.COMPR_SC
			caps.append(cap)

			cap = TS_SOUND_CAPABILITYSET()
			caps.append(cap)

			share_hdr = TS_SHARECONTROLHEADER()
			share_hdr.pduType = PDUTYPE.CONFIRMACTIVEPDU
			share_hdr.pduVersion = 1
			share_hdr.pduSource = self.__joined_channels['MCS'].channel_id

			msg = TS_CONFIRM_ACTIVE_PDU()
			msg.shareID = 0x103EA
			msg.originatorID = 1002
			for cap in caps:
				msg.capabilitySets.append(TS_CAPS_SET.from_capability(cap))
			
			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.handle_out_data(msg, sec_hdr, None, share_hdr, self.__joined_channels['MCS'].channel_id, False)
			data, err = await self.__joined_channels['MCS'].out_queue.get()
			if err is not None:
				raise err
			
			data = data[data_start_offset:]
			shc = TS_SHARECONTROLHEADER.from_bytes(data)
			if shc.pduType == PDUTYPE.DATAPDU:
				shd = TS_SHAREDATAHEADER.from_bytes(data)
				if shd.pduType2 == PDUTYPE2.SET_ERROR_INFO_PDU:
					# we got an error!
					res = TS_SET_ERROR_INFO_PDU.from_bytes(data)
					raise Exception('Server replied with error! Code: %s ErrName: %s' % (hex(res.errorInfoRaw), res.errorInfo.name))

				elif shd.pduType2 == PDUTYPE2.SYNCHRONIZE:
					# this is the expected data here
					res = TS_SYNCHRONIZE_PDU.from_bytes(data)
			
				else:
					raise Exception('Unexpected reply! %s' % shd.pduType2.name)
			else:
				raise Exception('Unexpected reply! %s' % shc.pduType.name)

			data_hdr = TS_SHAREDATAHEADER()
			data_hdr.shareID = 0x103EA
			data_hdr.streamID = STREAM_TYPE.MED
			data_hdr.pduType2 = PDUTYPE2.SYNCHRONIZE

			cli_sync = TS_SYNCHRONIZE_PDU()
			cli_sync.targetUser = self.__joined_channels['MCS'].channel_id
			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0
			
			await self.handle_out_data(cli_sync, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)

			data_hdr = TS_SHAREDATAHEADER()
			data_hdr.shareID = 0x103EA
			data_hdr.streamID = STREAM_TYPE.MED
			data_hdr.pduType2 = PDUTYPE2.CONTROL

			cli_ctrl = TS_CONTROL_PDU()
			cli_ctrl.action = CTRLACTION.COOPERATE
			cli_ctrl.grantId = 0
			cli_ctrl.controlId = 0

			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.handle_out_data(cli_ctrl, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)
			

			data_hdr = TS_SHAREDATAHEADER()
			data_hdr.shareID = 0x103EA
			data_hdr.streamID = STREAM_TYPE.MED
			data_hdr.pduType2 = PDUTYPE2.CONTROL

			cli_ctrl = TS_CONTROL_PDU()
			cli_ctrl.action = CTRLACTION.REQUEST_CONTROL
			cli_ctrl.grantId = 0
			cli_ctrl.controlId = 0

			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.handle_out_data(cli_ctrl, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)

			data_hdr = TS_SHAREDATAHEADER()
			data_hdr.shareID = 0x103EA
			data_hdr.streamID = STREAM_TYPE.MED
			data_hdr.pduType2 = PDUTYPE2.FONTLIST

			cli_font = TS_FONT_LIST_PDU()
			
			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.handle_out_data(cli_font, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)
			
			return True, None
		except Exception as e:
			return None, e

	async def __x224_reader(self):
		# recieves X224 packets and fastpath packets, performs decryption if necessary then dispatches each packet to 
		# the appropriate channel
		# gets activated when all channel setup is done
		# dont activate it before this!!!!
		
		try:
			self.__connection.packetizer.packetizer_control("X224")
			
			async for is_fastpath, response in self.__connection.read():
				#is_fastpath, response, err = await self._x224net.in_queue.get()
				#if err is not None:
				#	raise err

				if response is None:
					raise Exception('Server terminated the connection!')
				
				if is_fastpath is False:
					x = self._t125_per_codec.decode('DomainMCSPDU', response.data)
					if x[0] != 'sendDataIndication':
						#print('Unknown packet!')
						continue
					
					data = x[1]['userData']
					if data is not None:
						if self.cryptolayer is not None:
							sec_hdr = TS_SECURITY_HEADER1.from_bytes(data)
							if SEC_HDR_FLAG.ENCRYPT in sec_hdr.flags:
								orig_data = data[12:]
								data = self.cryptolayer.client_dec(orig_data)
								if SEC_HDR_FLAG.SECURE_CHECKSUM in sec_hdr.flags:
									mac = self.cryptolayer.calc_salted_mac(data, is_server=True)
								else:
									mac = self.cryptolayer.calc_mac(data)
								if mac != sec_hdr.dataSignature:
									print('ERROR! Signature mismatch! Printing debug data')
									print('Encrypted data: %s' % orig_data)
									print('Decrypted data: %s' % data)
									print('Original MAC  : %s' % sec_hdr.dataSignature)
									print('Calculated MAC: %s' % mac)
					await self.__channel_id_lookup[x[1]['channelId']].process_channel_data(data)
				else:
					#print('fastpath data in -> %s' % len(response))
					fpdu = TS_FP_UPDATE_PDU.from_bytes(response)
					if FASTPATH_SEC.ENCRYPTED in fpdu.flags:
						data = self.cryptolayer.client_dec(fpdu.fpOutputUpdates)
						if FASTPATH_SEC.SECURE_CHECKSUM in fpdu.flags:
							mac = self.cryptolayer.calc_salted_mac(data, is_server=True)
						else:
							mac = self.cryptolayer.calc_mac(data)
						if mac != fpdu.dataSignature:
							print('ERROR! Signature mismatch! Printing debug data')
							print('FASTPATH_SEC  : %s' % fpdu)
							print('Encrypted data: %s' % fpdu.fpOutputUpdates[:100])
							print('Decrypted data: %s' % data[:100])
							print('Original MAC  : %s' % fpdu.dataSignature)
							print('Calculated MAC: %s' % mac)
							raise Exception('Signature mismatch')
						fpdu.fpOutputUpdates = TS_FP_UPDATE.from_bytes(data)
					await self.__process_fastpath(fpdu)
		
		except asyncio.CancelledError:
			return None, None
		except Exception as e:
			traceback.print_exc()
			return None, e
		finally:
			await self.terminate()

	async def __process_fastpath(self, fpdu):
		# Fastpath was introduced to the RDP specs to speed up data transmission
		# by reducing 4 useless layers from the traffic.
		# Transmission on this channel starts immediately after connection sequence
		# mostly video and channel related info coming in from the server.
		# interesting note: it seems newer servers (>=win2016) only support this protocol of sending
		# high bandwith traffic. If you disable fastpath (during connection sequence) you won't
		# get images at all
		
		try:
			if fpdu.fpOutputUpdates.fragmentation != FASTPATH_FRAGMENT.SINGLE:
				print('WARNING! FRAGMENTATION IS NOT IMPLEMENTED! %s' % fpdu.fpOutputUpdates.fragmentation)
			if fpdu.fpOutputUpdates.updateCode == FASTPATH_UPDATETYPE.BITMAP:
				for bitmapdata in fpdu.fpOutputUpdates.update.rectangles:
					self.desktop_buffer_has_data = True
					res, image = RDP_VIDEO.from_bitmapdata(bitmapdata, self.iosettings.video_out_format)
					self.__desktop_buffer.paste(image, [res.x, res.y, res.x+res.width, res.y+res.height])
					await self.ext_out_queue.put(res)
			#else:
			#	#print(fpdu.fpOutputUpdates.updateCode)
			#	#if fpdu.fpOutputUpdates.updateCode == FASTPATH_UPDATETYPE.CACHED:
			#	#	print(fpdu.fpOutputUpdates)
			#	#if fpdu.fpOutputUpdates.updateCode not in [FASTPATH_UPDATETYPE.CACHED, FASTPATH_UPDATETYPE.POINTER]:
			#	#	print('notbitmap %s' % fpdu.fpOutputUpdates.updateCode.name)
		except Exception as e:
			# the decoder is not perfect yet, so it's better to keep this here...
			traceback.print_exc()
			return
	


	async def send_key_virtualkey(self, vk, is_pressed, is_extended, scancode_hint = None, modifiers = VK_MODIFIERS(0)):
		try:
			if vk in self.__vk_to_sc:
				scancode = self.__vk_to_sc[vk]
				is_extended = True
			else:
				scancode = scancode_hint
			return await self.send_key_scancode(scancode, is_pressed, is_extended)
		except Exception as e:
			traceback.print_exc()
			return None, e
	
	async def send_key_scancode(self, scancode, is_pressed, is_extended, modifiers = VK_MODIFIERS(0)):
		try:
			data_hdr = TS_SHAREDATAHEADER()
			data_hdr.shareID = 0x103EA
			data_hdr.streamID = STREAM_TYPE.MED
			data_hdr.pduType2 = PDUTYPE2.INPUT
			
			kbi = TS_KEYBOARD_EVENT()
			kbi.keyCode = scancode
			kbi.keyboardFlags = 0
			if is_pressed is False:
				kbi.keyboardFlags |= KBDFLAGS.RELEASE
			if is_extended is True or kbi.keyCode > 57000:
				kbi.keyboardFlags |= KBDFLAGS.EXTENDED
			clii_kb = TS_INPUT_EVENT.from_input(kbi)
			cli_input = TS_INPUT_PDU_DATA()
			cli_input.slowPathInputEvents.append(clii_kb)
			
			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.handle_out_data(cli_input, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)
				

		except Exception as e:
			traceback.print_exc()
			return None, e

	async def send_key_char(self, char, is_pressed):
		try:
			data_hdr = TS_SHAREDATAHEADER()
			data_hdr.shareID = 0x103EA
			data_hdr.streamID = STREAM_TYPE.MED
			data_hdr.pduType2 = PDUTYPE2.INPUT
			
			kbi = TS_UNICODE_KEYBOARD_EVENT()
			kbi.unicodeCode = char
			kbi.keyboardFlags = 0
			if is_pressed is False:
				kbi.keyboardFlags |= KBDFLAGS.RELEASE
			clii_kb = TS_INPUT_EVENT.from_input(kbi)
			cli_input = TS_INPUT_PDU_DATA()
			cli_input.slowPathInputEvents.append(clii_kb)

			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.handle_out_data(cli_input, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)
			return True, None

		except Exception as e:
			traceback.print_exc()
			return None, e

	async def send_mouse(self, button:MOUSEBUTTON, xPos:int, yPos:int, is_pressed:bool):
		try:
			if xPos < 0 or yPos < 0:
				return True, None
			data_hdr = TS_SHAREDATAHEADER()
			data_hdr.shareID = 0x103EA
			data_hdr.streamID = STREAM_TYPE.MED
			data_hdr.pduType2 = PDUTYPE2.INPUT
			
			mouse = TS_POINTER_EVENT()
			mouse.pointerFlags = 0
			if is_pressed is True:
				mouse.pointerFlags |= PTRFLAGS.DOWN
			if button == MOUSEBUTTON.MOUSEBUTTON_LEFT:
				mouse.pointerFlags |= PTRFLAGS.BUTTON1
			if button == MOUSEBUTTON.MOUSEBUTTON_RIGHT:
				mouse.pointerFlags |= PTRFLAGS.BUTTON2
			if button == MOUSEBUTTON.MOUSEBUTTON_MIDDLE:
				mouse.pointerFlags |= PTRFLAGS.BUTTON3
			if button == MOUSEBUTTON.MOUSEBUTTON_HOVER:
				# indicates a simple pointer update with no buttons pressed
				# sending this enables the mouse hover feel on the remote end
				mouse.pointerFlags |= PTRFLAGS.MOVE
			mouse.xPos = xPos
			mouse.yPos = yPos

			clii_mouse = TS_INPUT_EVENT.from_input(mouse)
					
			cli_input = TS_INPUT_PDU_DATA()
			cli_input.slowPathInputEvents.append(clii_mouse)

			sec_hdr = None
			if self.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

					
			await self.handle_out_data(cli_input, sec_hdr, data_hdr, None, self.__joined_channels['MCS'].channel_id, False)
		except Exception as e:
			traceback.print_exc()
			return None, e

	def get_desktop_buffer(self, encoding:VIDEO_FORMAT = VIDEO_FORMAT.PIL):
		"""Makes a copy of the current desktop buffer, converts it and returns the object"""
		try:
			image = copy.deepcopy(self.__desktop_buffer)
			if encoding == VIDEO_FORMAT.PIL:
				return image
			elif encoding == VIDEO_FORMAT.RAW:
				return image.tobytes()
			elif encoding == VIDEO_FORMAT.QT5:
				from PIL.ImageQt import ImageQt
				return ImageQt(image)
			elif encoding == VIDEO_FORMAT.PNG:
				img_byte_arr = io.BytesIO()
				image.save(img_byte_arr, format='PNG')
				return img_byte_arr.getvalue()
			else:
				raise ValueError('Output format of "%s" is not supported!' % encoding)
		except Exception as e:
			traceback.print_exc()
			return None, e
	
	async def get_current_clipboard_text(self):
		if 'cliprdr' not in self.__joined_channels:
			return None
		return await self.__joined_channels['cliprdr'].get_current_clipboard_text()

	async def set_current_clipboard_text(self, text:str):
		if 'cliprdr' not in self.__joined_channels:
			return None
		return await self.__joined_channels['cliprdr'].set_current_clipboard_text(text)
	
	async def add_vchannel(self, channelname, handler):
		if 'drdynvc' not in self.__joined_channels:
			raise Exception('Dynamic Virtual Channels are not enabled on this connection!')
		if channelname in self.__joined_channels['drdynvc'].defined_channels:
			raise Exception('Channel already defined!')
		self.__joined_channels['drdynvc'].defined_channels[channelname] = handler
	
	def get_vchannels(self):
		if 'drdynvc' not in self.__joined_channels:
			raise Exception('Dynamic Virtual Channels are not enabled on this connection!')
		return self.__joined_channels['drdynvc'].defined_channels
	
	async def __external_reader(self):
		# This coroutine handles keyboard/mouse etc input from the user
		# It wraps the data in it's appropriate format then dispatches it to the server
		try:
			while True:
				indata = await self.ext_in_queue.get()
				if indata is None:
					#signaling exit
					await self.terminate()
					return
				if indata.type == RDPDATATYPE.KEYSCAN:
					indata = cast(RDP_KEYBOARD_SCANCODE, indata)
					#right side control, altgr, and pause buttons still dont work well...
					#if indata.keyCode in [97]:
					#	await self.send_key_virtualkey('VK_RCONTROL', indata.is_pressed, indata.is_extended, scancode_hint=indata.keyCode)
					if indata.vk_code is not None:
						await self.send_key_virtualkey(indata.vk_code, indata.is_pressed, indata.is_extended, scancode_hint=indata.keyCode)
					else:
						await self.send_key_scancode(indata.keyCode, indata.is_pressed, indata.is_extended)
					
				elif indata.type == RDPDATATYPE.KEYUNICODE:
					indata = cast(RDP_KEYBOARD_UNICODE, indata)
					await self.send_key_char(indata.char, indata.is_pressed)

				elif indata.type == RDPDATATYPE.MOUSE:
					indata = cast(RDP_MOUSE, indata)
					await self.send_mouse(indata.button, indata.xPos, indata.yPos, indata.is_pressed)

				elif indata.type == RDPDATATYPE.CLIPBOARD_DATA_TXT:
					if 'cliprdr' not in self.__joined_channels:
						logger.debug('Got clipboard data but no clipboard channel setup!')
						continue
					await self.__joined_channels['cliprdr'].process_user_data(indata)

		except asyncio.CancelledError:
			return None, None

		except Exception as e:
			traceback.print_exc()
			await self.terminate()
			return None, e
	
	async def handle_out_data(self, dataobj, sec_hdr, datacontrol_hdr, sharecontrol_hdr, channel_id, is_fastpath):
		try:
			if is_fastpath is False:
				#print('Sending data on channel "%s(%s)"' % (self.name, self.channel_id))
				data = dataobj.to_bytes()
				hdrs = b''
				if sharecontrol_hdr is not None:
					sharecontrol_hdr.pduSource = channel_id
					sharecontrol_hdr.totalLength = len(data) + 6
					hdrs += sharecontrol_hdr.to_bytes()

				elif datacontrol_hdr is not None:
					datacontrol_hdr.shareControlHeader = TS_SHARECONTROLHEADER()
					datacontrol_hdr.shareControlHeader.pduType = PDUTYPE.DATAPDU
					datacontrol_hdr.shareControlHeader.pduSource = channel_id
					datacontrol_hdr.shareControlHeader.totalLength = len(data) + 24
					datacontrol_hdr.uncompressedLength = len(data) + 24 # since there is no compression implemented yet
					datacontrol_hdr.compressedType = 0
					datacontrol_hdr.compressedLength = 0
					hdrs += datacontrol_hdr.to_bytes()
				if sec_hdr is not None:
					sec_hdr = typing.cast(TS_SECURITY_HEADER, sec_hdr)
					if SEC_HDR_FLAG.ENCRYPT in sec_hdr.flags:
						#print('PacketCount: %s' % self.connection.cryptolayer.PacketCount)
						data = hdrs+data
							

						if self.cryptolayer.use_encrypted_mac is True:
							checksum = self.cryptolayer.calc_salted_mac(data)
							sec_hdr.flags |= SEC_HDR_FLAG.SECURE_CHECKSUM
						else:
							checksum = self.cryptolayer.calc_mac(data)
							
						# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/9791c9e2-e5be-462f-8c23-3404c4af63b3
						enc_data = self.cryptolayer.client_enc(data)
							
						data = checksum + enc_data
						hdrs = sec_hdr.to_bytes()
					else:
						hdrs += sec_hdr.to_bytes()
				userdata = hdrs + data
				data_wrapper = {
					'initiator': self._initiator,
					'channelId': channel_id,
					'dataPriority': 'high',
					'segmentation': (b'\xc0', 2),
					'userData': userdata
				}
				userdata_wrapped = self._t125_per_codec.encode('DomainMCSPDU', ('sendDataRequest', data_wrapper))
				await self._x224net.write(userdata_wrapped)
				
			else:
				raise NotImplementedError("Fastpath output is not yet implemented")

		except Exception as e:
			traceback.print_exc()
			await self.terminate()
			return None, e
		
	
async def amain():
	try:
		from aardwolf.commons.factory import RDPConnectionFactory
		from aardwolf.commons.iosettings import RDPIOSettings
		from aardwolf.extensions.RDPEDYC.channel import RDPEDYCChannel

		iosettings = RDPIOSettings()
		iosettings.channels.append(RDPEDYCChannel)
		url = 'rdp+ntlm-password://TEST\\Administrator:Passw0rd!1@10.10.10.102'
		rdpurl = RDPConnectionFactory.from_url(url, iosettings)
		conn = rdpurl.get_connection(iosettings)
		_, err = await conn.connect()
		if err is not None:
			raise err
		
		while True:
			data = await conn.ext_out_queue.get()
			#print(data)
	except Exception as e:
		traceback.print_exc()

	

if __name__ == '__main__':
	asyncio.run(amain())


import enum
import io
from aardwolf.extensions.RDPEDYC.protocol import DYNVC_CMD, dynvc_header_to_bytes, dynvc_header_from_buff

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/f82105dd-0abd-4126-a61b-41a7909e974f

class DYNVC_SOFT_SYNC_FLAG(enum.IntFlag):
    TCP_FLUSHED = 0x01
    CHANNEL_LIST_PRESENT = 0x02

class DYNVC_SOFT_SYNC_REQUEST:
    def __init__(self):
        self.cbid:int = None
        self.sp:int   = 0
        self.cmd:DYNVC_CMD = DYNVC_CMD.SOFT_SYNC_REQUEST
        self.Pad:int   = 0
        self.Length:int   = None
        self.Flags:DYNVC_SOFT_SYNC_FLAG = None
        self.NumberOfTunnels:int = None
        self.SoftSyncChannelLists:List[DYNVC_SOFT_SYNC_CHANNEL_LIST] = []

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_SOFT_SYNC_REQUEST.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_SOFT_SYNC_REQUEST()
        msg.cbid, msg.sp, msg.cmd = dynvc_header_from_buff(buff)
        msg.Pad = buff.read(1)[0]
        msg.Length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
        msg.Flags = DYNVC_SOFT_SYNC_FLAG(int.from_bytes(buff.read(2), byteorder='little', signed=False))
        msg.NumberOfTunnels = int.from_bytes(buff.read(2), byteorder='little', signed=False)
        for _ in range(msg.NumberOfTunnels):
            msg.SoftSyncChannelLists.append(DYNVC_SOFT_SYNC_CHANNEL_LIST.from_buffer(buff))
        return msg

    def to_bytes(self):
        t = b''
        t += dynvc_header_to_bytes(self.cbid, self.sp, self.cmd)
        t += self.Pad.to_bytes(1, byteorder='little', signed=False)
        t += self.Length.to_bytes(4, byteorder='little', signed=False)
        t += self.Flags.to_bytes(2, byteorder='little', signed=False)
        t += len(self.SoftSyncChannelLists).to_bytes(2, byteorder='little', signed=False)
        for ch in self.SoftSyncChannelLists:
            t += ch.to_bytes()
        return t

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/f82105dd-0abd-4126-a61b-41a7909e974f

from typing import List

class DYNVC_TUNNELTYPE(enum.Enum):
    UDPFECR = 0x00000001 #RDP-UDP Forward Error Correction (FEC) multitransport tunnel ([MS-RDPEMT] section 1.3).
    UDPFECL = 0x00000003 #RDP-UDP FEC lossy multitransport tunnel ([MS-RDPEMT] section 1.3).


class DYNVC_SOFT_SYNC_CHANNEL_LIST:
    def __init__(self):
        self.TunnelType:DYNVC_TUNNELTYPE = None
        self.NumberOfDVCs:int   = None
        self.ListOfDVCIds:List[int] = []

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_SOFT_SYNC_CHANNEL_LIST.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_SOFT_SYNC_CHANNEL_LIST()
        msg.TunnelType = DYNVC_TUNNELTYPE(int.from_bytes(buff.read(4), byteorder='little', signed=False))
        msg.NumberOfDVCs = int.from_bytes(buff.read(2), byteorder='little', signed=False)
        for _ in range(msg.NumberOfDVCs):
            msg.ListOfDVCIds.append(int.from_bytes(buff.read(4), byteorder='little', signed=False))
        return msg

    def to_bytes(self):
        t = b''
        t += self.TunnelType.value.to_bytes(4, byteorder='little', signed=False)
        t += len(self.ListOfDVCIds).to_bytes(2, byteorder='little', signed=False)
        for channelid in self.ListOfDVCIds:
            t += channelid.to_bytes(4, byteorder='little', signed=False)
        return t

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/7d120558-b35d-4b78-81e8-bac2cf081bd7
class DYNVC_SOFT_SYNC_RESPONSE:
    def __init__(self):
        self.cbid:int = None
        self.sp:int   = 0
        self.cmd:DYNVC_CMD = DYNVC_CMD.SOFT_SYNC_RESPONSE
        self.Pad:int  = 0
        self.NumberOfTunnels:int = None
        self.TunnelsToSwitch:List[DYNVC_TUNNELTYPE] = []

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_SOFT_SYNC_RESPONSE.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_SOFT_SYNC_RESPONSE()
        msg.cbid, msg.sp, msg.cmd = dynvc_header_from_buff(buff)
        msg.Pad = buff.read(1)[0]
        msg.NumberOfTunnels = int.from_bytes(buff.read(2), byteorder='little', signed=False)
        for _ in range(msg.NumberOfTunnels):
            msg.TunnelsToSwitch.append(DYNVC_TUNNELTYPE(int.from_bytes(buff.read(4), byteorder='little', signed=False)))
        return msg

    def to_bytes(self):
        t = b''
        t += dynvc_header_to_bytes(self.cbid, self.sp, self.cmd)
        t += self.Pad.to_bytes(1, byteorder='little', signed=False)
        t += len(self.NumberOfTunnels).to_bytes(2, byteorder='little', signed=False)
        for ch in self.TunnelsToSwitch:
            t += ch.value.to_bytes(4, byteorder='little', signed=False)
        return t


import io
import enum


class DRAWFLAG(enum.IntFlag):
	ALLOW_DYNAMIC_COLOR_FIDELITY = 0x02 #Indicates support for lossy compression of 32 bpp bitmaps by reducing color-fidelity on a per-pixel basis ([MS-RDPEGDI] section 3.1.9.1.4).
	ALLOW_COLOR_SUBSAMPLING = 0x04 #Indicates support for chroma subsampling when compressing 32 bpp bitmaps ([MS-RDPEGDI] section 3.1.9.1.3).
	ALLOW_SKIP_ALPHA = 0x08 #Indicates that the client supports the removal of the alpha-channel when compressing 32 bpp bitmaps. In this case the alpha is assumed to be 0xFF, meaning the bitmap is opaque.
	UNUSED_FLAG = 0x10 #An unused flag that MUST be ignored by the client if it is present in the server-to-client Bitmap Capability Set.



class TS_BITMAP_CAPABILITYSET:
	def __init__(self):
		self.preferredBitsPerPixel:int = None
		self.receive1BitPerPixel:bool = True
		self.receive4BitsPerPixel:bool = True
		self.receive8BitsPerPixel:bool = True
		self.desktopWidth:int = None
		self.desktopHeight:int = None
		self.pad2octets:int = 0
		self.desktopResizeFlag:bool = False
		self.bitmapCompressionFlag:bool = True
		self.highColorFlags:int = 0
		self.drawingFlags:DRAWFLAG = DRAWFLAG.UNUSED_FLAG
		self.multipleRectangleSupport:bool = True
		self.pad2octetsB:int = 0

	def to_bytes(self):
		t = self.preferredBitsPerPixel.to_bytes(2, byteorder='little', signed = False)
		t += int(self.receive1BitPerPixel).to_bytes(2, byteorder='little', signed = False)
		t += int(self.receive4BitsPerPixel).to_bytes(2, byteorder='little', signed = False)
		t += int(self.receive8BitsPerPixel).to_bytes(2, byteorder='little', signed = False)
		t += self.desktopWidth.to_bytes(2, byteorder='little', signed = False)
		t += self.desktopHeight.to_bytes(2, byteorder='little', signed = False)
		t += self.pad2octets.to_bytes(2, byteorder='little', signed = False)
		t += int(self.desktopResizeFlag).to_bytes(2, byteorder='little', signed = False)
		t += int(self.bitmapCompressionFlag).to_bytes(2, byteorder='little', signed = False)
		t += self.highColorFlags.to_bytes(1, byteorder='little', signed = False)
		t += self.drawingFlags.to_bytes(1, byteorder='little', signed = False)
		t += int(self.multipleRectangleSupport).to_bytes(2, byteorder='little', signed = False)
		t += int(self.pad2octetsB).to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAP_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAP_CAPABILITYSET()
		msg.preferredBitsPerPixel = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.receive1BitPerPixel = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.receive4BitsPerPixel = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.receive8BitsPerPixel = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.desktopWidth = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.desktopHeight = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pad2octets = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.desktopResizeFlag = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.bitmapCompressionFlag = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.highColorFlags = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.drawingFlags = DRAWFLAG(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		msg.multipleRectangleSupport = bool(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.pad2octetsB = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_BITMAP_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


"""
RFB protocol implementattion, client side.
Override RFBClient and RFBFactory in your application.
See vncviewer.py for an example.
Reference:
http://www.realvnc.com/docs/rfbproto.pdf
(C) 2003 cliechti@gmx.net
MIT License
"""

KEY_BackSpace = 0xff08
KEY_Tab = 0xff09
KEY_Return = 0xff0d
KEY_Escape = 0xff1b
KEY_Insert = 0xff63
KEY_Delete = 0xffff
KEY_Home = 0xff50
KEY_End = 0xff57
KEY_PageUp = 0xff55
KEY_PageDown = 0xff56
KEY_Left = 0xff51
KEY_Up = 0xff52
KEY_Right = 0xff53
KEY_Down = 0xff54
KEY_F1 = 0xffbe
KEY_F2 = 0xffbf
KEY_F3 = 0xffc0
KEY_F4 = 0xffc1
KEY_F5 = 0xffc2
KEY_F6 = 0xffc3
KEY_F7 = 0xffc4
KEY_F8 = 0xffc5
KEY_F9 = 0xffc6
KEY_F10 = 0xffc7
KEY_F11 = 0xffc8
KEY_F12 = 0xffc9
KEY_F13 = 0xFFCA
KEY_F14 = 0xFFCB
KEY_F15 = 0xFFCC
KEY_F16 = 0xFFCD
KEY_F17 = 0xFFCE
KEY_F18 = 0xFFCF
KEY_F19 = 0xFFD0
KEY_F20 = 0xFFD1
KEY_ShiftLeft = 0xffe1
KEY_ShiftRight = 0xffe2
KEY_ControlLeft = 0xffe3
KEY_ControlRight = 0xffe4
KEY_MetaLeft = 0xffe7
KEY_MetaRight = 0xffe8
KEY_AltLeft = 0xffe9
KEY_AltRight = 0xffea

KEY_Scroll_Lock = 0xFF14
KEY_Sys_Req = 0xFF15
KEY_Num_Lock = 0xFF7F
KEY_Caps_Lock = 0xFFE5
KEY_Pause = 0xFF13
KEY_Super_L = 0xFFEB
KEY_Super_R = 0xFFEC
KEY_Hyper_L = 0xFFED
KEY_Hyper_R = 0xFFEE
KEY_Menu_R = 0xFF67

KEY_KP_0 = 0xFFB0
KEY_KP_1 = 0xFFB1
KEY_KP_2 = 0xFFB2
KEY_KP_3 = 0xFFB3
KEY_KP_4 = 0xFFB4
KEY_KP_5 = 0xFFB5
KEY_KP_6 = 0xFFB6
KEY_KP_7 = 0xFFB7
KEY_KP_8 = 0xFFB8
KEY_KP_9 = 0xFFB9
KEY_KP_Add = 0xFFAB
KEY_KP_Subtract = 0xFFAD
KEY_KP_Multiply = 0xFFAA
KEY_KP_Decimal = 0xFFAE
KEY_KP_Divide = 0xFFAF
KEY_KP_Enter = 0xFF8D
KEY_Print = 0xFF61, 




import io
from aardwolf.extensions.RDPEDYC.protocol import DYNVC_CMD, dynvc_header_to_bytes, dynvc_header_from_buff


# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpedyc/c02dfd21-ccbc-4254-985b-3ef6dd115dec
class DYNVC_CLOSE:
    def __init__(self):
        self.cbid:int = 3
        self.sp:int   = 0
        self.cmd:DYNVC_CMD = DYNVC_CMD.CLOSE
        self.ChannelId:int = None

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_CLOSE.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_CLOSE()
        msg.cbid, msg.sp, msg.cmd = dynvc_header_from_buff(buff, cbid_mod=True)
        msg.ChannelId = int.from_bytes(buff.read(msg.cbid), byteorder='little', signed=False)
        return msg

    def to_bytes(self):
        if self.cbid is None:
            self.cbid = 4
        t = b''
        t += dynvc_header_to_bytes(self.cbid, self.sp, self.cmd, cbid_mod=True)
        t += self.ChannelId.to_bytes(self.cbid, byteorder='little', signed=False)
        return t